#!/usr/bin/env python

__status__ = "Development"

import matplotlib
import numpy as np

matplotlib.use("agg")

import warnings

import matplotlib.cm as cm
import matplotlib.pyplot as plt

warnings.warn(
    "Replaced PyCDT usage of DefectPlotter objects with plotting ability of"
    "DefectPhaseDiagram from pymatgen.analysis.defects.thermodynamics\n"
    "Will remove DefectPlotter with Version 2.5 of PyCDT.",
    DeprecationWarning,
)


class DefectPlotter(object):
    """
    Class performing all the typical plots from a defects study
    """

    def __init__(self, dpd):
        """
        Args:
            dpd: DefectPhaseDiagram object from pymatgen.analysis.defects.thermodynamics
        """

        self._dpd = dpd

    def get_plot_form_energy(
        self,
        mu_elts,
        xlim=None,
        ylim=None,
        ax_fontsize=1.3,
        lg_fontsize=1.0,
        lg_position=None,
    ):
        """
        Formation energy vs Fermi energy plot
        Args:
            mu_elts:
                a dictionnary of {Element:value} giving the chemical
                potential of each element
            xlim:
                Tuple (min,max) giving the range of the x (fermi energy) axis
            ylim:
                Tuple (min,max) giving the range for the formation energy axis
            ax_fontsize:
                float  multiplier to change axis label fontsize
            lg_fontsize:
                float  multiplier to change legend label fontsize
            lg_position:
                Tuple (horizontal-position, vertical-position) giving the position
                to place the legend.
                Example: (0.5,-0.75) will likely put it below the x-axis.

        Returns:
            a matplotlib object

        """
        if xlim is None:
            xlim = (-0.5, self._dpd.band_gap + 1.5)
        xy = {}
        lower_cap = -100.0
        upper_cap = 100.0
        for defnom, def_tl in self._dpd.transition_level_map.items():
            xy[defnom] = [[], []]
            if def_tl:
                org_x = list(def_tl.keys())  # list of transition levels
                org_x.sort()  # sorted with lowest first

                # establish lower x-bound
                first_charge = max(def_tl[org_x[0]])
                for chg_ent in self._dpd.stable_entries[defnom]:
                    if chg_ent.charge == first_charge:
                        form_en = chg_ent.formation_energy(
                            chemical_potentials=mu_elts, fermi_level=lower_cap
                        )
                xy[defnom][0].append(lower_cap)
                xy[defnom][1].append(form_en)

                # iterate over stable charge state transitions
                for fl in org_x:
                    charge = min(def_tl[fl])
                    for chg_ent in self._dpd.stable_entries[defnom]:
                        if chg_ent.charge == charge:
                            form_en = chg_ent.formation_energy(
                                chemical_potentials=mu_elts, fermi_level=fl
                            )
                    xy[defnom][0].append(fl)
                    xy[defnom][1].append(form_en)

                # establish upper x-bound
                last_charge = min(def_tl[org_x[-1]])
                for chg_ent in self._dpd.stable_entries[defnom]:
                    if chg_ent.charge == last_charge:
                        form_en = chg_ent.formation_energy(
                            chemical_potentials=mu_elts, fermi_level=upper_cap
                        )
                xy[defnom][0].append(upper_cap)
                xy[defnom][1].append(form_en)
            else:
                # no transition - just one stable charge
                chg_ent = self._dpd.stable_entries[defnom][0]
                for x_extrem in [lower_cap, upper_cap]:
                    xy[defnom][0].append(x_extrem)
                    xy[defnom][1].append(
                        chg_ent.formation_energy(
                            chemical_potentials=mu_elts, fermi_level=x_extrem
                        )
                    )

        width, height = 12, 8
        plt.clf()
        if len(xy) <= 8:
            colors = cm.Dark2(np.linspace(0, 1, len(xy)))
        else:
            colors = cm.gist_rainbow(np.linspace(0, 1, len(xy)))

        # plot formation energy lines
        for_legend = []
        for cnt, defnom in enumerate(xy.keys()):
            plt.plot(xy[defnom][0], xy[defnom][1], linewidth=3, color=colors[cnt])
            for_legend.append(self._dpd.stable_entries[defnom][0].copy())

        # plot transtition levels
        for cnt, defnom in enumerate(xy.keys()):
            x_trans, y_trans = [], []
            for x_val, chargeset in self._dpd.transition_level_map[defnom].items():
                x_trans.append(x_val)
                for chg_ent in self._dpd.stable_entries[defnom]:
                    if chg_ent.charge == chargeset[0]:
                        form_en = chg_ent.formation_energy(
                            chemical_potentials=mu_elts, fermi_level=x_val
                        )
                y_trans.append(form_en)
            if len(x_trans):
                plt.plot(
                    x_trans,
                    y_trans,
                    marker="*",
                    color=colors[cnt],
                    markersize=12,
                    fillstyle="full",
                )

        # get latex-like legend titles
        legends_txt = []
        for dfct in for_legend:
            flds = dfct.name.split("_")
            if "Vac" == flds[0]:
                base = "$Vac"
                sub_str = "_{" + flds[1] + "}$"
            elif "Sub" == flds[0]:
                flds = dfct.name.split("_")
                base = "$" + flds[1]
                sub_str = "_{" + flds[3] + "}$"
            elif "Int" == flds[0]:
                base = "$" + flds[1]
                # sub_str = '_{i_{'+','.join(flds[2])+'}}$'
                sub_str = "_{inter}$"
            else:
                base = dfct.name
                sub_str = ""

            legends_txt.append(base + sub_str)

        if not lg_position:
            plt.legend(legends_txt, fontsize=lg_fontsize * width, loc=0)
        else:
            plt.legend(
                legends_txt,
                fontsize=lg_fontsize * width,
                ncol=3,
                loc="lower center",
                bbox_to_anchor=lg_position,
            )

        if ylim is not None:
            plt.ylim(ylim)

        plt.xlim(xlim)
        plt.plot(
            [xlim[0], xlim[1]], [0, 0], "k-"
        )  # black dashed line for Eformation = 0
        plt.axvline(
            x=0.0, linestyle="--", color="k", linewidth=3
        )  # black dashed lines for gap edges
        plt.axvline(x=self._dpd.band_gap, linestyle="--", color="k", linewidth=3)
        plt.xlabel("Fermi energy (eV)", size=ax_fontsize * width)
        plt.ylabel("Defect Formation\nEnergy (eV)", size=ax_fontsize * width)

        return plt


class StructureRelaxPlotter(object):
    """
    This class plots movement of atomic sites as a function of radius

    relaxation_data is list of [distance to defect, distance atom moved,
                                index in structure, percentage contribution to total relaxation]

    """

    def __init__(self, relaxation_data, sampling_radius):
        rd = relaxation_data[:]
        rd.sort()
        self.relaxation_data = np.array(rd)
        self.sampling_radius = sampling_radius

    def plot(self, title=""):
        plt.figure()
        plt.clf()
        fig, ax1 = plt.subplots()

        ax1.set_xlabel("Radius from Defect ($\AA$)", fontsize=20)

        ax1.plot(
            self.relaxation_data[:, 0],
            self.relaxation_data[:, 1],
            "k",
            marker="o",
            linestyle="--",
        )

        ax2 = ax1.twinx()
        ax2.plot(
            self.relaxation_data[:, 0],
            self.relaxation_data[:, 3],
            "k",
            marker="o",
            linestyle="--",
        )

        tmpx = [self.sampling_radius, max(self.relaxation_data[:, 0])]
        max_fill = max(self.relaxation_data[:, 3])
        min_fill = min(self.relaxation_data[:, 3])
        plt.fill_between(
            tmpx,
            min_fill,
            max_fill,
            facecolor="red",
            alpha=0.15,
            label="delocalization region",
        )

        ax1.set_ylabel("Relaxation amount ($\AA$)", color="b", fontsize=15)
        ax2.set_ylabel("Percentage of total relaxation (%)\n", color="r", fontsize=15)
        plt.legend(loc=0)

        plt.title(str(title) + " Atomic Relaxation", fontsize=20)

        return plt


class SingleParticlePlotter(object):
    """
    This class plots single particle KS wavefunction as a function of radiusfrom defect

    relaxation_data is list of [distance to defect, distance atom moved,
                                index in structure, percentage contribution to total relaxation]

    """

    def __init__(self, defect_ks_delocal_data):
        self.defect_ks_delocal_data = defect_ks_delocal_data
        self.nspin = len(defect_ks_delocal_data["localized_band_indices"])
        lbl_dict = defect_ks_delocal_data["localized_band_indices"]
        self.localized_bands = set(
            [band_index for spin_list in lbl_dict.values() for band_index in spin_list]
        )
        print("Localized KS wavefunction bands are {}".format(self.localized_bands))

    def plot(self, bandnum, title=""):
        if bandnum not in self.localized_bands:
            raise ValueError("{} is not in {}".format(bandnum, self.localized_bands))

        plt.figure()
        plt.clf()
        fig, ax1 = plt.subplots()

        final_out = self.defect_ks_delocal_data["followup_wf_parse"][bandnum]
        dat = final_out["0"]["rad_dist_data"]["tot"]

        ax1.set_xlabel("Radius from Defect ($\AA$)")
        ax1.plot(dat[0], dat[1], "b")
        ax2 = ax1.twinx()
        ax2.plot(dat[0], 100.0 * np.array(dat[2]), "r")

        if self.nspin == 2:
            dat = final_out["1"]["rad_dist_data"]["tot"]
            ax1.plot(dat[0], -1.0 * np.array(dat[1]), "b-")
            ax2.plot(dat[0], -100.0 * np.array(dat[2]), "r-")

        ax1.set_ylabel("Occupation", color="b")
        ax2.set_ylabel("Percentage of electron contained\n", color="r")
        plt.title(str(title) + " KS band index " + str(bandnum))

        return plt



# coding: utf-8

from __future__ import division

import glob
import os
import unittest

from monty.json import MontyDecoder
from monty.serialization import loadfn
from monty.tempfile import ScratchDir
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Incar, Kpoints, Poscar, Potcar

from doped.pycdt.utils.vasp import *

__status__ = "Development"

file_loc = os.path.abspath(os.path.join(__file__, "..", "..", "..", "..", "test_files"))


class PotcarSingleModTest(unittest.TestCase):
    """
    This test is applicable for the specific case where POTCAR files are not
    organized according to The Materials Project directory layout, but
    in the default layout.
    """

    def setUp(self):
        pass

    @unittest.expectedFailure
    def test_from_symbol_and_functional(self):
        try:
            potcar = PotcarSingleMod.from_symbol_and_functional("Ni")
        except:
            potcar = None

        self.assertIsNotNone(potcar)


class PotcarModTest(unittest.TestCase):
    def setUp(self):
        pass

    @unittest.expectedFailure
    def test_set_symbols(self):
        try:
            potcar = PotcarMod(symbols=["Ni", "O"])
        except:
            potcar = None

        self.assertIsNotNone(potcar)


class DefectRelaxTest(unittest.TestCase):
    def setUp(self):
        self.structure = Structure.from_file(os.path.join(file_loc, "POSCAR_Cr2O3"))
        self.user_settings = loadfn(os.path.join(file_loc, "test_vasp_settings.yaml"))
        self.path = "Cr2O3"
        self.neutral_def_incar_min = {
            "LVHAR": True,
            "ISYM": 0,
            "ISMEAR": 0,
            "ISIF": 2,
            "ISPIN": 2,
        }
        self.def_keys = ["EDIFF", "EDIFFG", "IBRION"]

    def test_neutral_defect_incar(self):
        drs = DefectRelaxSet(self.structure)
        self.assertTrue(self.neutral_def_incar_min.items() <= drs.incar.items())
        self.assertTrue(set(self.def_keys).issubset(drs.incar))

    @unittest.expectedFailure
    def test_charged_defect_incar(self):
        drs = DefectRelaxSet(self.structure, charge=1)
        self.assertIn("NELECT", drs.incar)
        self.assertTrue(self.neutral_def_incar_min.items() <= drs.incar.items())
        self.assertTrue(set(self.def_keys).issubset(drs.incar))

    def test_user_settings_defect_incar(self):
        user_incar_settings = {"EDIFF": 1e-8, "EDIFFG": 0.1, "ENCUT": 720}
        drs = DefectRelaxSet(
            self.structure, charge=1, user_incar_settings=user_incar_settings
        )
        self.assertTrue(self.neutral_def_incar_min.items() <= drs.incar.items())
        self.assertTrue(set(self.def_keys).issubset(drs.incar))
        self.assertEqual(drs.incar["ENCUT"], 720)
        self.assertEqual(drs.incar["EDIFF"], 1e-8)
        self.assertEqual(drs.incar["EDIFFG"], 0.1)


class MakeVaspDefectFilesTest(unittest.TestCase):
    def setUp(self):
        self.defects = loadfn(os.path.join(file_loc, "Cr2O3_defects.json"))
        self.user_settings = loadfn(os.path.join(file_loc, "test_vasp_settings.yaml"))
        self.path = "Cr2O3"
        self.neutral_def_incar_min = {
            "LVHAR": True,
            "ISYM": 0,
            "ISMEAR": 0,
            "ISIF": 2,
            "ISPIN": 2,
        }
        self.def_keys = ["EDIFF", "EDIFFG", "IBRION"]
        self.bulk_incar = {
            "LVHAR": True,
            "ISYM": 0,
            "ISMEAR": 0,
            "IBRION": -1,
            "ISPIN": 2,
        }
        self.bulk_keys = ["EDIFF"]

    def test_neutral_defect_incar(self):
        with ScratchDir("."):
            make_vasp_defect_files(self.defects, self.path)
            cr_def_path = glob.glob(os.path.join(self.path, "vac*Cr"))[0]
            incar_loc = os.path.join(cr_def_path, "charge_0")
            incar = Incar.from_file(os.path.join(incar_loc, "INCAR"))
            self.assertTrue(self.neutral_def_incar_min.items() <= incar.items())
            self.assertTrue(set(self.def_keys).issubset(incar))

    def test_charged_defect_incar(self):
        with ScratchDir("."):
            make_vasp_defect_files(self.defects, self.path)
            cr_def_path = glob.glob(os.path.join(self.path, "vac*Cr"))[0]
            incar_loc = os.path.join(cr_def_path, "charge_-1")
            try:
                potcar = Potcar.from_file(os.path.join(incar_loc, "POTCAR"))
            except:
                potcar = None
            if potcar:
                incar = Incar.from_file(os.path.join(incar_loc, "INCAR"))
                self.assertIsNotNone(incar.pop("NELECT", None))
                self.assertTrue(self.neutral_def_incar_min.items() <= incar.items())
                self.assertTrue(set(self.def_keys).issubset(incar))

    def test_bulk_incar(self):
        with ScratchDir("."):
            make_vasp_defect_files(self.defects, self.path)
            incar_loc = os.path.join(self.path, "bulk")
            incar = Incar.from_file(os.path.join(incar_loc, "INCAR"))
            self.assertTrue(self.bulk_incar.items() <= incar.items())
            self.assertTrue(set(self.bulk_keys).issubset(incar))

    def test_kpoints(self):
        with ScratchDir("."):
            make_vasp_defect_files(self.defects, self.path)
            kpoints_loc = os.path.join(self.path, "bulk")
            kpoints = Kpoints.from_file(os.path.join(kpoints_loc, "KPOINTS"))

    def test_poscar(self):
        with ScratchDir("."):
            make_vasp_defect_files(self.defects, self.path)
            poscar_loc = os.path.join(self.path, "bulk")
            poscar = Poscar.from_file(os.path.join(poscar_loc, "POSCAR"))
            self.assertTrue(
                poscar.structure.matches(self.defects["bulk"]["supercell"]["structure"])
            )

    def test_user_settings_bulk(self):
        user_settings = loadfn(os.path.join(file_loc, "test_vasp_settings.yaml"))
        with ScratchDir("."):
            make_vasp_defect_files(self.defects, self.path, user_settings=user_settings)
            incar_loc = os.path.join(self.path, "bulk")
            incar = Incar.from_file(os.path.join(incar_loc, "INCAR"))
            self.assertTrue(self.bulk_incar.items() <= incar.items())
            self.assertTrue(set(self.bulk_keys).issubset(incar))
            self.assertEqual(incar["ENCUT"], 620)

    def test_hse_settings(self):
        pass


class MakeVaspDielectricFilesTest(unittest.TestCase):
    def setUp(self):
        self.structure = Structure.from_file(os.path.join(file_loc, "POSCAR_Cr2O3"))
        self.user_settings = loadfn(os.path.join(file_loc, "test_vasp_settings.yaml"))
        self.path = "dielectric"
        self.dielectric_min = {"LPEAD": True, "LEPSILON": True, "IBRION": 8}
        self.keys = ["EDIFF"]

    def test_dielectric_files(self):
        with ScratchDir("."):
            make_vasp_dielectric_files(self.structure, self.path)
            incar = Incar.from_file(os.path.join(self.path, "INCAR"))
            self.assertTrue(self.dielectric_min.items() <= incar.items())
            self.assertTrue(set(self.keys).issubset(incar))

    def test_user_modified_dielectric_files(self):
        with ScratchDir("."):
            make_vasp_dielectric_files(
                self.structure, self.path, user_settings=self.user_settings
            )
            incar = Incar.from_file(os.path.join(self.path, "INCAR"))
            self.assertTrue(self.dielectric_min.items() <= incar.items())
            self.assertTrue(set(self.keys).issubset(incar))
            self.assertEqual(incar["EDIFF"], 5e-7)
            self.assertEqual(incar["ENCUT"], 620)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
"""
Parses the computed data from VASP defect calculations.
"""
# from __future__ import unicode_literals
from __future__ import division

import glob
import logging
import os
import warnings

import numpy as np
from monty.json import MontyDecoder
from monty.serialization import dumpfn, loadfn
from pymatgen.analysis.defects.core import (
    DefectEntry,
    Interstitial,
    Substitution,
    Vacancy,
)
from pymatgen.analysis.defects.defect_compatibility import DefectCompatibility
from pymatgen.analysis.defects.utils import TopographyAnalyzer
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core import PeriodicSite, Structure
from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.ext.matproj import MPRester
from pymatgen.io.vasp.inputs import Potcar, UnknownPotcarWarning
from pymatgen.io.vasp.outputs import Locpot, Outcar, Poscar, Vasprun
from pymatgen.util.coord import pbc_diff

from doped.pycdt.core import _chemical_potentials

angstrom = "\u212B"  # unicode symbol for angstrom to print in strings

# globally ignore these POTCAR warnings
warnings.filterwarnings("ignore", category=UnknownPotcarWarning)
warnings.filterwarnings("ignore", message="No POTCAR file with matching TITEL fields")
warnings.filterwarnings("ignore", message="Ignoring unknown variable type")

# until updated from pymatgen==2022.7.25 :
warnings.filterwarnings(
    "ignore", message="Using `tqdm.autonotebook.tqdm` in notebook mode"
)
warnings.filterwarnings(
    "ignore", message="`np.int` is a deprecated alias for the builtin `int`"
)
warnings.filterwarnings("ignore", message="Use get_magnetic_symmetry()")


def custom_formatwarning(msg, *args, **kwargs):
    """Reformat warnings to just print the warning message"""
    return f"{msg}\n"


warnings.formatwarning = custom_formatwarning


def convert_cd_to_de(cd, b_cse):
    """
    As of pymatgen v2.0, ComputedDefect objects were deprecated in favor
    of DefectEntry objects in pymatgen.analysis.defects.core
    This function takes a ComputedDefect (either as a dict or object) and
    converts it into a DefectEntry object in order to handle legacy
    PyCDT creation within the current paradigm of PyCDT.

    :param cd (dict or ComputedDefect object): ComputedDefect as an object or as a dictionary
    :params b_cse (dict or ComputedStructureEntry object): ComputedStructureEntry of bulk entry
        associated with the ComputedDefect.
    :return: de (DefectEntry): Resulting DefectEntry object
    """
    if isinstance(cd, dict):
        cd = cd.as_dict()
    if isinstance(b_cse, dict):
        b_cse = b_cse.as_dict()

    bulk_sc_structure = Structure.from_dict(b_cse["structure"])

    # modify defect_site as required for Defect object, confirming site exists in bulk structure
    site_cls = cd["site"]
    defect_site = PeriodicSite.from_dict(site_cls)
    def_nom = cd["name"].lower()
    if "sub_" in def_nom or "as_" in def_nom:
        # modify site object for substitution site of Defect object
        site_cls["species"][0]["element"] = cd["name"].split("_")[2]
        defect_site = PeriodicSite.from_dict(site_cls)

    poss_deflist = sorted(
        bulk_sc_structure.get_sites_in_sphere(
            defect_site.coords, 0.2, include_index=True
        ),
        key=lambda x: x[1],
    )
    if len(poss_deflist) != 1:
        raise ValueError(
            "ComputedDefect to DefectEntry conversion failed. "
            "Could not determine periodic site position in bulk supercell."
        )

    # create defect object
    if "vac_" in def_nom:
        defect_obj = Vacancy(bulk_sc_structure, defect_site, charge=cd["charge"])
    elif "as_" in def_nom or "sub_" in def_nom:
        defect_obj = Substitution(bulk_sc_structure, defect_site, charge=cd["charge"])
    elif "int_" in def_nom:
        defect_obj = Interstitial(bulk_sc_structure, defect_site, charge=cd["charge"])
    else:
        raise ValueError(f"Could not recognize defect type for {cd['name']}")

    # assign proper energy and parameter metadata
    uncorrected_energy = cd["entry"]["energy"] - b_cse["energy"]
    def_path = os.path.split(cd["entry"]["data"]["locpot_path"])[0]
    bulk_path = os.path.split(b_cse["data"]["locpot_path"])[0]
    p = {
        "defect_path": def_path,
        "bulk_path": bulk_path,
        "encut": cd["entry"]["data"]["encut"],
    }

    de = DefectEntry(defect_obj, uncorrected_energy, parameters=p)

    return de


def get_vasprun(vasprun_path, **kwargs):
    """Read the vasprun.xml(.gz) file as a pymatgen Vasprun object"""
    vasprun_path = str(vasprun_path)  # convert to string if Path object
    warnings.filterwarnings(
        "ignore", category=UnknownPotcarWarning
    )  # Ignore POTCAR warnings when loading vasprun.xml
    # pymatgen assumes the default PBE with no way of changing this within get_vasprun())
    warnings.filterwarnings(
        "ignore", message="No POTCAR file with matching TITEL fields"
    )
    if os.path.exists(vasprun_path):
        vasprun = Vasprun(vasprun_path)
        read_vasprun_path = vasprun_path
    elif os.path.exists(vasprun_path + ".gz", **kwargs):
        vasprun = Vasprun(vasprun_path + ".gz", **kwargs)
        read_vasprun_path = vasprun_path + ".gz"
    else:
        raise FileNotFoundError(
            f"""vasprun.xml(.gz) not found at {vasprun_path}(.gz). Needed for parsing defect 
            calculations."""
        )
    return vasprun, read_vasprun_path


def get_locpot(locpot_path):
    """Read the LOCPOT(.gz) file as a pymatgen Locpot object"""
    locpot_path = str(locpot_path)  # convert to string if Path object
    if os.path.exists(locpot_path):
        locpot = Locpot.from_file(locpot_path)
    elif os.path.exists(locpot_path + ".gz"):
        locpot = Locpot.from_file(locpot_path + ".gz")
    else:
        raise FileNotFoundError(
            f"""LOCPOT(.gz) not found at {locpot_path}(.gz). Needed for calculating the 
            Freysoldt (FNV) image charge corrections."""
        )
    return locpot


def get_outcar(outcar_path):
    """Read the OUTCAR(.gz) file as a pymatgen Outcar object"""
    outcar_path = str(outcar_path)  # convert to string if Path object
    if os.path.exists(outcar_path):
        outcar = Outcar(outcar_path)
    elif os.path.exists(outcar_path + ".gz"):
        outcar = Outcar(outcar_path + ".gz")
    else:
        raise FileNotFoundError(
            f"""OUTCAR(.gz) not found at {outcar_path}(.gz). Needed for calculating the Kumagai (
            eFNV) image charge corrections."""
        )
    return outcar


def get_defect_type_and_composition_diff(bulk, defect):
    """Get the difference in composition between a bulk structure and a defect structure.
    Contributed by Dr. Alex Ganose (@ Imperial Chemistry) and refactored for extrinsic species
    """
    bulk_comp = bulk.composition.get_el_amt_dict()
    defect_comp = defect.composition.get_el_amt_dict()

    composition_diff = {
        element: int(defect_amount - bulk_comp.get(element, 0))
        for element, defect_amount in defect_comp.items()
        if int(defect_amount - bulk_comp.get(element, 0)) != 0
    }

    if len(composition_diff) == 1 and list(composition_diff.values())[0] == 1:
        defect_type = "interstitial"
    elif len(composition_diff) == 1 and list(composition_diff.values())[0] == -1:
        defect_type = "vacancy"
    elif len(composition_diff) == 2:
        defect_type = "substitution"
    else:
        raise RuntimeError(
            "Could not determine defect type from composition difference of bulk "
            "and defect structures."
        )

    return defect_type, composition_diff


def get_defect_site_idxs_and_unrelaxed_structure(
    bulk, defect, defect_type, composition_diff, unique_tolerance=1
):
    """Get the defect site and unrelaxed structure.
    Contributed by Dr. Alex Ganose (@ Imperial Chemistry) and refactored for extrinsic species
    """
    if defect_type == "substitution":
        old_species = [el for el, amt in composition_diff.items() if amt == -1][0]
        new_species = [el for el, amt in composition_diff.items() if amt == 1][0]

        bulk_new_species_coords = np.array(
            [site.frac_coords for site in bulk if site.specie.name == new_species]
        )
        defect_new_species_coords = np.array(
            [site.frac_coords for site in defect if site.specie.name == new_species]
        )
        defect_new_species_idx = np.array(
            [defect.index(site) for site in defect if site.specie.name == new_species]
        )

        if bulk_new_species_coords.size > 0:  # intrinsic substitution
            # find coords of new species in defect structure, taking into account periodic
            # boundaries
            distance_matrix = np.linalg.norm(
                pbc_diff(bulk_new_species_coords[:, None], defect_new_species_coords),
                axis=2,
            )
            site_matches = distance_matrix.argmin(axis=1)

            if len(np.unique(site_matches)) != len(site_matches):
                raise RuntimeError(
                    "Could not uniquely determine site of new species in defect structure"
                )

            defect_site_idx = list(
                set(np.arange(len(defect_new_species_coords), dtype=int))
                - set(site_matches)
            )[0]

        else:  # extrinsic substitution
            defect_site_idx = 0

        defect_coords = defect_new_species_coords[defect_site_idx]

        # Get the site index of the defect that was used in the VASP calculation
        defect_site_idx = defect_new_species_idx[defect_site_idx]

        # now find the closest old_species site in the bulk structure to the defect site
        # again, make sure to use periodic boundaries
        bulk_old_species_coords = np.array(
            [site.frac_coords for site in bulk if site.specie.name == old_species]
        )
        distances = np.linalg.norm(
            pbc_diff(bulk_old_species_coords, defect_coords), axis=1
        )
        original_site_idx = distances.argmin()

        # if there are any other matches with a distance within unique_tolerance of the located
        # site then unique matching failed
        if (
            len(distances[distances < distances[original_site_idx] * unique_tolerance])
            > 1
        ):
            raise RuntimeError(
                "Could not uniquely determine site of old species in bulk structure"
            )

        # currently, original_site_idx is indexed with respect to the old species only.
        # Need to get the index in the full structure
        bulk_coords = np.array([s.frac_coords for s in bulk])
        bulk_site_idx = np.linalg.norm(
            pbc_diff(bulk_coords, bulk_old_species_coords[original_site_idx]), axis=1
        ).argmin()

        # create unrelaxed defect structure
        unrelaxed_defect_structure = bulk.copy()
        unrelaxed_defect_structure.remove_sites([bulk_site_idx])
        # Place defect in same location as output from DFT
        unrelaxed_defect_structure.insert(
            defect_site_idx, new_species, bulk_coords[bulk_site_idx]
        )

    elif defect_type == "vacancy":
        old_species = list(composition_diff.keys())[0]

        bulk_old_species_coords = np.array(
            [site.frac_coords for site in bulk if site.specie.name == old_species]
        )
        defect_old_species_coords = np.array(
            [site.frac_coords for site in defect if site.specie.name == old_species]
        )

        # make sure to do take into account periodic boundaries
        distance_matrix = np.linalg.norm(
            pbc_diff(bulk_old_species_coords[:, None], defect_old_species_coords),
            axis=2,
        )
        site_matches = distance_matrix.argmin(axis=0)

        if len(np.unique(site_matches)) != len(site_matches):
            raise RuntimeError(
                "Could not uniquely determine site of vacancy in defect structure"
            )

        original_site_idx = list(
            set(np.arange(len(bulk_old_species_coords), dtype=int)) - set(site_matches)
        )[0]

        # currently, original_site_idx is indexed with respect to the old species only.
        # Need to get the index in the full structure
        bulk_coords = np.array([s.frac_coords for s in bulk])
        bulk_site_idx = np.linalg.norm(
            pbc_diff(bulk_coords, bulk_old_species_coords[original_site_idx]), axis=1
        ).argmin()

        # create unrelaxed defect structure
        unrelaxed_defect_structure = bulk.copy()
        unrelaxed_defect_structure.remove_sites([bulk_site_idx])
        defect_site_idx = None

    elif defect_type == "interstitial":
        new_species = list(composition_diff.keys())[0]

        bulk_new_species_coords = np.array(
            [site.frac_coords for site in bulk if site.specie.name == new_species]
        )
        defect_new_species_coords = np.array(
            [site.frac_coords for site in defect if site.specie.name == new_species]
        )
        defect_new_species_idx = np.array(
            [defect.index(site) for site in defect if site.specie.name == new_species]
        )

        if bulk_new_species_coords.size > 0:  # intrinsic interstitial
            # make sure to take into account periodic boundaries
            distance_matrix = np.linalg.norm(
                pbc_diff(bulk_new_species_coords[:, None], defect_new_species_coords),
                axis=2,
            )
            site_matches = distance_matrix.argmin(axis=1)

            if len(np.unique(site_matches)) != len(site_matches):
                raise RuntimeError(
                    "Could not uniquely determine site of interstitial in defect structure"
                )

            defect_site_idx = list(
                set(np.arange(len(defect_new_species_coords), dtype=int))
                - set(site_matches)
            )[0]

        else:  # extrinsic interstitial
            defect_site_idx = 0

        defect_site_coords = defect_new_species_coords[defect_site_idx]

        # Get the site index of the defect that was used in the VASP calculation
        defect_site_idx = defect_new_species_idx[defect_site_idx]

        # create unrelaxed defect structure
        unrelaxed_defect_structure = bulk.copy()
        # Place defect in same location as output from DFT
        unrelaxed_defect_structure.insert(
            defect_site_idx, new_species, defect_site_coords
        )
        bulk_site_idx = None

    else:
        raise ValueError(f"Invalid defect type: {defect_type}")

    return (
        bulk_site_idx,
        defect_site_idx,
        unrelaxed_defect_structure,
    )


class SingleDefectParser:
    # TODO: Given this is our recommended workflow, should streamline this similar to in
    #  `vasp_input.py`, where it tries to run the freysoldt/kumagai loader (chosen by an
    #  optional imput parameter, if not set check if OUTCAR/LOCPOT present), throw warning if not
    #  successful, then `get_stdrd_metadata()`, `get_bulk_gap_data()` and `run_compatibility()` (
    #  again throwing warnings for any of these that don't work) – so replacing 4+ function calls
    #  with just one.
    # TODO: Should also add check the bulk and defect KPOINTS/INCAR/POTCAR/POSCAR (size) settings
    #  are compatible, and throw warning if not.
    # TODO: Add `check_defects_compatibility()` function that checks the bulk and defect
    #  KPOINTS/INCAR/POTCAR/POSCAR (size) settings for all defects in the supplied defect_dict
    #  are compatible, if not throw warnings and say what the differences are. Should recommend
    #  using this in the example notebook if a user has parsed the defects individually (rather
    #  than with the single looping function described below):
    # TODO: Add a function that loops over all the defects in a directory (with `defect_dir = .`,
    #  and `subfolder = vasp_ncl` options) and parses them all, returning a dictionary of defect
    #  entries, with the defect name as the key. (i.e. doing the loop in the example notebook).
    #  Show both this function and the individual function calls in the example notebook. Benefit
    #  of this one is that we can then auto-run `check_defects_compatibility()` at the end of
    #  parsing the full defects dict.
    def __init__(
        self,
        defect_entry,
        compatibility=DefectCompatibility(
            plnr_avg_var_tol=0.01,
            plnr_avg_minmax_tol=0.3,
            atomic_site_var_tol=0.025,
            atomic_site_minmax_tol=0.3,
            tot_relax_tol=5.0,
            defect_tot_relax_tol=5.0,
            use_bandfilling=False,  # don't include bandfilling by default
            use_bandedgeshift=False,  # don't include band edge shift by default
        ),
        defect_vr=None,
        bulk_vr=None,
    ):
        """
        Parse a defect object using features that resemble that of a standard
        DefectBuilder object (emmet), but without the requirement of atomate.
        Also allows for use of DefectCompatibility object within pymatgen

        :param defect_entry (DefectEntry): DefectEntry of interest (using the bulk supercell as
        bulk_structure)
            NOTE: to make use of methods within the class, bulk_path and and defect_path
            must exist within the defect_entry parameters class.
        :param compatibility (DefectCompatibility): Compatibility class instance for
            performing charge correction compatibility analysis on defect entry.
        :param defect_vr (Vasprun):
        :param bulk_vr (Vasprun):

        """
        self.defect_entry = defect_entry
        self.compatibility = compatibility
        self.defect_vr = defect_vr
        self.bulk_vr = bulk_vr

    @staticmethod
    def from_paths(
        path_to_defect,
        path_to_bulk,
        dielectric,
        defect_charge,
        mpid=None,
        compatibility=DefectCompatibility(
            plnr_avg_var_tol=0.01,
            plnr_avg_minmax_tol=0.3,
            atomic_site_var_tol=0.025,
            atomic_site_minmax_tol=0.3,
            tot_relax_tol=5.0,
            defect_tot_relax_tol=5.0,
            use_bandfilling=False,  # don't include bandfilling by default
            use_bandedgeshift=False,  # don't include band edge shift by default
        ),
        initial_defect_structure=None,
    ):
        """
        Identify defect object based on file paths. Minimal parsing performing for
        instantiating the SingleDefectParser class.

        Args:
        path_to_defect (str): path to defect folder of interest (with vasprun.xml(.gz))
        path_to_bulk (str): path to bulk folder of interest (with vasprun.xml(.gz))
        dielectric (float or int or 3x1 matrix or 3x3 matrix):
            ionic + static contributions to dielectric constant
        defect_charge (int):  charge of defect
        mpid (str):  Materials Project ID of bulk structure
        compatibility (DefectCompatibility): Compatibility class instance for
            performing charge correction compatibility analysis on defect entry.
        initial_defect_structure (str):  Path to the unrelaxed defect structure,
            if structure matching with the relaxed defect structure(s) fails.

        Return:
            Instance of the SingleDefectParser class.
        """
        # check if dielectric in required 3x3 matrix format
        if not isinstance(dielectric, (float, int)):
            dielectric = np.array(dielectric)
            if dielectric.shape == (3,):
                dielectric = np.diag(dielectric)
            elif dielectric.shape != (3, 3):
                raise ValueError(
                    f"Dielectric constant must be a float/int or a 3x1 matrix or 3x3 matrix, "
                    f"got type {type(dielectric)} and shape {dielectric.shape}"
                )

        parameters = {
            "bulk_path": path_to_bulk,
            "defect_path": path_to_defect,
            "dielectric": dielectric,
            "mpid": mpid,
        }

        # add bulk simple properties
        bulk_vr, bulk_vr_path = get_vasprun(os.path.join(path_to_bulk, "vasprun.xml"))
        bulk_energy = bulk_vr.final_energy
        bulk_sc_structure = bulk_vr.initial_structure.copy()

        # add defect simple properties
        defect_vr, defect_vr_path = get_vasprun(
            os.path.join(path_to_defect, "vasprun.xml")
        )
        defect_energy = defect_vr.final_energy
        # Can specify initial defect structure (to help PyCDT find the defect site if
        # multiple relaxations were required, else use from defect relaxation OUTCAR:
        if initial_defect_structure:
            initial_defect_structure = Poscar.from_file(
                initial_defect_structure
            ).structure.copy()
        else:
            initial_defect_structure = defect_vr.initial_structure.copy()

        # Add initial defect structure to parameters, so it can be pulled later on
        # (eg. for Kumagai loader)
        parameters["initial_defect_structure"] = initial_defect_structure

        # identify defect site, structural information, and create defect object
        try:
            def_type, comp_diff = get_defect_type_and_composition_diff(
                bulk_sc_structure, initial_defect_structure
            )
        except RuntimeError as exc:
            raise ValueError(
                "Could not identify defect type from number of sites in structure: "
                f"{len(bulk_sc_structure)} in bulk vs. {len(initial_defect_structure)} in defect?"
            ) from exc

        bulk_site_idx = None
        defect_site_idx = None
        unrelaxed_defect_structure = None
        # Try automatic defect site detection - this gives us the "unrelaxed" defect structure
        try:
            (
                bulk_site_idx,
                defect_site_idx,
                unrelaxed_defect_structure,
            ) = get_defect_site_idxs_and_unrelaxed_structure(
                bulk_sc_structure, initial_defect_structure, def_type, comp_diff
            )
        except RuntimeError as exc:
            # if auto site-matching failed, try use transformation.json
            # The goal is to find the `defect_site_idx` or `defect_site_idx` based on the
            # tranformation.
            transformation_path = os.path.join(path_to_defect, "transformation.json")
            if not os.path.exists(transformation_path):  # try next folder up
                orig_transformation_path = transformation_path
                transformation_path = os.path.join(
                    os.path.dirname(os.path.normpath(path_to_defect)),
                    "transformation.json",
                )
                if os.path.exists(transformation_path):
                    print(
                        f"No transformation file found at {orig_transformation_path}, but found "
                        f"one at {transformation_path}. Using this for defect parsing."
                    )

            if os.path.exists(transformation_path):
                tf = loadfn(transformation_path)
                site = tf["defect_supercell_site"]
                if def_type == "vacancy":
                    poss_deflist = sorted(
                        bulk_sc_structure.get_sites_in_sphere(
                            site.coords, 0.2, include_index=True
                        ),
                        key=lambda x: x[1],
                    )
                    searched = "bulk_supercell"
                    if poss_deflist:
                        bulk_site_idx = poss_deflist[0][2]
                else:
                    poss_deflist = sorted(
                        initial_defect_structure.get_sites_in_sphere(
                            site.coords, 2.5, include_index=True
                        ),
                        key=lambda x: x[1],
                    )
                    searched = "initial_defect_structure"
                    if poss_deflist:
                        defect_site_idx = poss_deflist[0][2]
                if not poss_deflist:
                    raise ValueError(
                        f"{transformation_path} specified defect site {site}, but could not find "
                        f"it in {searched}. Abandoning parsing."
                    ) from exc
                if poss_deflist[0][1] > 1:
                    site_matched_defect = poss_deflist[0]  # pymatgen Neighbor object
                    offsite_warning = (
                        f"Site-matching has determined {site_matched_defect.species} at "
                        f"{site_matched_defect.coords} as the defect site, located "
                        f"{site_matched_defect.nn_distance:.2f} {angstrom} from its initial "
                        f"position. This may incur small errors in the charge correction."
                    )
                    warnings.warn(message=offsite_warning)

            else:
                raise RuntimeError(
                    f"Could not identify {def_type} defect site in defect structure. "
                    f"Try supplying the initial defect structure to "
                    f"SingleDefectParser.from_paths(), or making sure the doped "
                    f"transformation.json files are in the defect directory."
                ) from exc

        if def_type == "vacancy":
            defect_site = bulk_sc_structure[bulk_site_idx]
        else:
            if unrelaxed_defect_structure:
                defect_site = unrelaxed_defect_structure[defect_site_idx]
            else:
                defect_site = initial_defect_structure[defect_site_idx]

        if unrelaxed_defect_structure:
            if def_type == "interstitial":
                # get closest Voronoi site in bulk supercell to final interstitial site as this is
                # likely to be the initial interstitial site
                try:
                    struc_and_node_dict = loadfn("./bulk_voronoi_nodes.json")
                    if not StructureMatcher(
                        stol=0.05,
                        primitive_cell=False,
                        scale=False,
                        attempt_supercell=False,
                        allow_subset=False,
                    ).fit(struc_and_node_dict["bulk_supercell"], bulk_sc_structure):
                        warnings.warn(
                            "Previous bulk_voronoi_nodes.json detected, but does not "
                            "match current bulk supercell. Recalculating Voronoi nodes."
                        )
                        raise FileNotFoundError

                    voronoi_frac_coords = struc_and_node_dict["Voronoi nodes"]

                except FileNotFoundError:  # first time parsing
                    topography = TopographyAnalyzer(
                        bulk_sc_structure,
                        bulk_sc_structure.symbol_set,
                        [],
                        check_volume=False,
                    )
                    topography.cluster_nodes()
                    topography.remove_collisions()
                    voronoi_frac_coords = [
                        site.frac_coords for site in topography.vnodes
                    ]
                    struc_and_node_dict = {
                        "bulk_supercell": bulk_sc_structure,
                        "Voronoi nodes": voronoi_frac_coords,
                    }
                    dumpfn(
                        struc_and_node_dict, "./bulk_voronoi_nodes.json"
                    )  # for efficient
                    # parsing of multiple defects at once
                    print(
                        "Saving parsed Voronoi sites (for interstitial site-matching) to "
                        "bulk_voronoi_sites.json to speed up future parsing."
                    )

                closest_node_frac_coords = min(
                    voronoi_frac_coords,
                    key=lambda node: defect_site.distance_and_image_from_frac_coords(
                        node
                    )[0],
                )
                int_site = unrelaxed_defect_structure[defect_site_idx]
                unrelaxed_defect_structure.remove_sites([defect_site_idx])
                unrelaxed_defect_structure.insert(
                    defect_site_idx,  # Place defect at same position as in DFT calculation
                    int_site.species_string,
                    closest_node_frac_coords,
                    coords_are_cartesian=False,
                    validate_proximity=True,
                )
                defect_site = unrelaxed_defect_structure[defect_site_idx]

            # Use the unrelaxed_defect_structure to fix the initial defect structure
            initial_defect_structure = reorder_unrelaxed_structure(
                unrelaxed_defect_structure, initial_defect_structure
            )
            parameters["initial_defect_structure"] = initial_defect_structure
            parameters["unrelaxed_defect_structure"] = unrelaxed_defect_structure
        else:
            warnings.warn(
                "Cannot determine the unrelaxed `initial_defect_structure`. Please ensure the "
                "`initial_defect_structure` is indeed unrelaxed."
            )

        for_monty_defect = {
            "@module": "pymatgen.analysis.defects.core",
            "@class": def_type.capitalize(),
            "charge": defect_charge,
            "structure": bulk_sc_structure,
            "defect_site": defect_site,
        }
        defect = MontyDecoder().process_decoded(for_monty_defect)

        if unrelaxed_defect_structure:
            # only do StructureMatcher test if unrelaxed structure exists
            test_defect_structure = defect.generate_defect_structure()
            if not StructureMatcher(
                stol=0.25,
                primitive_cell=False,
                scale=False,
                attempt_supercell=False,
                allow_subset=False,
            ).fit(test_defect_structure, unrelaxed_defect_structure):
                # NOTE: this does not insure that cartesian coordinates or indexing are identical
                raise ValueError(
                    "Error in defect object matching! Unrelaxed structure (1st below) "
                    "does not match pymatgen defect.generate_defect_structure() "
                    f"(2nd below):\n{unrelaxed_defect_structure}"
                    f"\n{test_defect_structure}"
                )

        defect_entry = DefectEntry(
            defect, defect_energy - bulk_energy, corrections={}, parameters=parameters
        )

        return SingleDefectParser(
            defect_entry,
            compatibility=compatibility,
            defect_vr=defect_vr,
            bulk_vr=bulk_vr,
        )

    def freysoldt_loader(self, bulk_locpot=None):
        """
        Load metadata required for performing Freysoldt correction
        requires "bulk_path" and "defect_path" to be loaded to DefectEntry parameters dict.
        Can read gunzipped "LOCPOT.gz" files as well.

        Args:
            bulk_locpot (Locpot): Add bulk Locpot object for expedited parsing.
                If None, will load from file path variable bulk_path
        Return:
            bulk_locpot object for reuse by another defect entry (for expedited parsing)
        """
        if not self.defect_entry.charge:
            # dont need to load locpots if charge is zero
            return None

        if not bulk_locpot:
            bulk_locpot_path = os.path.join(
                self.defect_entry.parameters["bulk_path"], "LOCPOT"
            )
            bulk_locpot = get_locpot(bulk_locpot_path)

        def_locpot_path = os.path.join(
            self.defect_entry.parameters["defect_path"], "LOCPOT"
        )
        def_locpot = get_locpot(def_locpot_path)

        axis_grid = [def_locpot.get_axis_grid(i) for i in range(3)]
        bulk_planar_averages = [bulk_locpot.get_average_along_axis(i) for i in range(3)]
        defect_planar_averages = [
            def_locpot.get_average_along_axis(i) for i in range(3)
        ]

        self.defect_entry.parameters.update(
            {
                "axis_grid": axis_grid,
                "bulk_planar_averages": bulk_planar_averages,
                "defect_planar_averages": defect_planar_averages,
                "defect_frac_sc_coords": self.defect_entry.site.frac_coords,
            }
        )
        if "unrelaxed_defect_structure" in self.defect_entry.parameters:
            self.defect_entry.parameters.update(
                {
                    "initial_defect_structure": self.defect_entry.parameters[
                        "unrelaxed_defect_structure"
                    ],
                }
            )

        return bulk_locpot

    # noinspection DuplicatedCode
    def kumagai_loader(self, bulk_outcar=None):
        """
        Load metadata required for performing Kumagai correction
        requires "bulk_path" and "defect_path" to be loaded to DefectEntry parameters dict.

        Args:
            bulk_outcar (Outcar): Add bulk Outcar object for expedited parsing.
                If None, will load from file path variable bulk_path
        Return:
            bulk_outcar object for reuse by another defect entry (for expedited parsing)
        """
        if not self.defect_entry.charge:
            # dont need to load outcars if charge is zero
            return None

        if not bulk_outcar:
            bulk_outcar_path = os.path.join(
                self.defect_entry.parameters["bulk_path"], "OUTCAR"
            )
            bulk_outcar = get_outcar(bulk_outcar_path)
        else:
            bulk_outcar_path = "`bulk_outcar`"

        def_outcar_path = os.path.join(
            self.defect_entry.parameters["defect_path"], "OUTCAR"
        )
        def_outcar = get_outcar(def_outcar_path)

        bulk_atomic_site_averages = bulk_outcar.electrostatic_potential
        defect_atomic_site_averages = def_outcar.electrostatic_potential
        if not bulk_atomic_site_averages:
            raise ValueError(
                f"Unable to parse atomic core potentials from bulk `OUTCAR` at "
                f"{bulk_outcar_path}. This can happen if `ICORELEVEL` was not set "
                f"to 0 (= default) in the `INCAR`, or if the calculation was "
                f"finished prematurely with a `STOPCAR`. The Kumagai charge "
                f"correction cannot be computed without this data!"
            )

        if not defect_atomic_site_averages:
            raise ValueError(
                f"Unable to parse atomic core potentials from defect `OUTCAR` at "
                f"{def_outcar_path}. This can happen if `ICORELEVEL` was not set "
                f"to 0 (= default) in the `INCAR`, or if the calculation was "
                f"finished prematurely with a `STOPCAR`. The Kumagai charge "
                f"correction cannot be computed without this data!"
            )

        bulk_structure = self.defect_entry.bulk_structure
        bulksites = [site.frac_coords for site in bulk_structure]

        defect_structure = self.defect_entry.parameters["initial_defect_structure"]
        initsites = [site.frac_coords for site in defect_structure]

        distmatrix = bulk_structure.lattice.get_all_distances(
            bulksites, initsites
        )  # first index of this list is bulk index
        min_dist_with_index = [
            [
                min(distmatrix[bulk_index]),
                int(bulk_index),
                int(distmatrix[bulk_index].argmin()),
            ]
            for bulk_index in range(len(distmatrix))
        ]  # list of [min dist, bulk ind, defect ind]

        site_matching_indices = []
        if isinstance(self.defect_entry.defect, (Vacancy, Interstitial)):
            for mindist, bulk_index, defect_index in min_dist_with_index:
                if mindist < 0.5:
                    site_matching_indices.append([bulk_index, defect_index])

        elif isinstance(self.defect_entry.defect, Substitution):
            for mindist, bulk_index, defect_index in min_dist_with_index:
                species_match = (
                    bulk_structure[bulk_index].specie
                    == defect_structure[defect_index].specie
                )
                if mindist < 0.5 and species_match:
                    site_matching_indices.append([bulk_index, defect_index])

        self.defect_entry.parameters.update(
            {
                "bulk_atomic_site_averages": bulk_atomic_site_averages,
                "defect_atomic_site_averages": defect_atomic_site_averages,
                "site_matching_indices": site_matching_indices,
                "defect_frac_sc_coords": self.defect_entry.site.frac_coords,
            }
        )

        return bulk_outcar

    def get_stdrd_metadata(self):
        if not self.bulk_vr:
            path_to_bulk = self.defect_entry.parameters["bulk_path"]
            self.bulk_vr, bulk_vr_path = get_vasprun(
                os.path.join(path_to_bulk, "vasprun.xml")
            )

        if not self.defect_vr:
            path_to_defect = self.defect_entry.parameters["defect_path"]
            self.defect_vr, defect_vr_path = get_vasprun(
                os.path.join(path_to_defect, "vasprun.xml")
            )

        # standard bulk metadata
        bulk_energy = self.bulk_vr.final_energy
        bulk_sc_structure = self.bulk_vr.initial_structure
        self.defect_entry.parameters.update(
            {"bulk_energy": bulk_energy, "bulk_sc_structure": bulk_sc_structure}
        )

        # standard run metadata
        run_metadata = {}
        run_metadata.update(
            {
                "defect_incar": self.defect_vr.incar,
                "bulk_incar": self.bulk_vr.incar,
                "defect_kpoints": self.defect_vr.kpoints,
                "bulk_kpoints": self.bulk_vr.kpoints,
            }
        )
        run_metadata.update(
            {
                "incar_calctype_summary": {
                    k: self.defect_vr.incar.get(k, None)
                    if self.defect_vr.incar.get(k) not in ["None", "False", False]
                    else None
                    for k in [
                        "LHFCALC",
                        "HFSCREEN",
                        "IVDW",
                        "LUSE_VDW",
                        "LDAU",
                        "METAGGA",
                    ]
                }
            }
        )
        run_metadata.update(
            {
                "potcar_summary": {
                    "pot_spec": [
                        potelt["titel"] for potelt in self.defect_vr.potcar_spec
                    ],
                    "pot_labels": self.defect_vr.potcar_spec,
                    "pot_type": self.defect_vr.run_type,
                }
            }
        )

        self.defect_entry.parameters.update({"run_metadata": run_metadata.copy()})

        # standard defect run metadata
        self.defect_entry.parameters.update(
            {
                "final_defect_structure": self.defect_vr.final_structure,
                "defect_energy": self.defect_vr.final_energy,
            }
        )

        # grab defect energy and eigenvalue information for band filling and localization analysis
        eigenvalues = {
            spincls.value: eigdict.copy()
            for spincls, eigdict in self.defect_vr.eigenvalues.items()
        }
        kpoint_weights = self.defect_vr.actual_kpoints_weights[:]
        self.defect_entry.parameters.update(
            {"eigenvalues": eigenvalues, "kpoint_weights": kpoint_weights}
        )

    def get_bulk_gap_data(self, no_MP=True, actual_bulk_path=None):
        """Get bulk gap data from Materials Project or from local OUTCAR file.

        Args:
            no_MP (bool): If True, will not query MP for bulk gap data. (Default: True)
            actual_bulk_path (str): Path to bulk OUTCAR file for determining the band gap. If
                the VBM/CBM occur at reciprocal space points not included in the bulk supercell
                calculation, you should use this tag to point to a bulk bandstructure calculation
                instead. If None, will use self.defect_entry.parameters["bulk_path"].
        """
        if not self.bulk_vr:
            path_to_bulk = self.defect_entry.parameters["bulk_path"]
            self.bulk_vr, bulk_vr_path = get_vasprun(
                os.path.join(path_to_bulk, "vasprun.xml")
            )

        bulk_sc_structure = self.bulk_vr.initial_structure
        mpid = self.defect_entry.parameters["mpid"]

        if not mpid and not no_MP:
            try:
                with MPRester() as mp:
                    tmp_mplist = mp.get_entries_in_chemsys(
                        list(bulk_sc_structure.symbol_set)
                    )
                mplist = [
                    ment.entry_id
                    for ment in tmp_mplist
                    if ment.composition.reduced_composition
                    == bulk_sc_structure.composition.reduced_composition
                ]
            except Exception as exc:
                raise ValueError(
                    f"Error with querying MPRester for"
                    f" {bulk_sc_structure.composition.reduced_formula}:"
                ) from exc

            mpid_fit_list = []
            for trial_mpid in mplist:
                with MPRester() as mp:
                    mpstruct = mp.get_structure_by_material_id(trial_mpid)
                if StructureMatcher(
                    primitive_cell=True,
                    scale=False,
                    attempt_supercell=True,
                    allow_subset=False,
                ).fit(bulk_sc_structure, mpstruct):
                    mpid_fit_list.append(trial_mpid)

            if len(mpid_fit_list) == 1:
                mpid = mpid_fit_list[0]
                print(f"Single mp-id found for bulk structure:{mpid}.")
            elif len(mpid_fit_list) > 1:
                num_mpid_list = [int(mp.split("" - "")[1]) for mp in mpid_fit_list]
                num_mpid_list.sort()
                mpid = "mp-" + str(num_mpid_list[0])
                print(
                    f"Multiple mp-ids found for bulk structure:{mpid_fit_list}\nWill use lowest "
                    f"number mpid for bulk band structure = {mpid}."
                )
            else:
                print(
                    "Could not find bulk structure in MP database after tying the following "
                    f"list:\n{mplist}"
                )
                mpid = None

        vbm, cbm, bandgap = None, None, None
        gap_parameters = {}
        if mpid is not None and not no_MP:
            print(f"Using user-provided mp-id for bulk structure: {mpid}.")
            with MPRester() as mp:
                bs = mp.get_bandstructure_by_material_id(mpid)
            if bs:
                cbm = bs.get_cbm()["energy"]
                vbm = bs.get_vbm()["energy"]
                bandgap = bs.get_band_gap()["energy"]
                gap_parameters.update(
                    {"MP_gga_BScalc_data": bs.get_band_gap().copy()}
                )  # contains gap kpt transition

        if (
            vbm is None
            or bandgap is None
            or cbm is None
            or no_MP
            or not actual_bulk_path
        ):
            if mpid and bandgap is None:
                print(
                    f"WARNING: Mpid {mpid} was provided, but no bandstructure entry currently "
                    "exists for it. \nReverting to use of bulk supercell calculation for band "
                    "edge extrema."
                )
            if mpid and no_MP:
                print(
                    f"Mpid {mpid} was provided, but `no_MP` flag was set to True. \n"
                    "Reverting to use of bulk supercell calculation for band edge extrema."
                )

            gap_parameters.update(
                {"MP_gga_BScalc_data": None}
            )  # to signal no MP BS is used
            bandgap, cbm, vbm, _ = self.bulk_vr.eigenvalue_band_properties

        if actual_bulk_path:
            print(f"Using actual bulk path: {actual_bulk_path}")
            actual_bulk_vr, actual_bulk_vr_path = get_vasprun(
                os.path.join(actual_bulk_path, "vasprun.xml")
            )
            bandgap, cbm, vbm, _ = actual_bulk_vr.eigenvalue_band_properties

        gap_parameters.update({"mpid": mpid, "cbm": cbm, "vbm": vbm, "gap": bandgap})
        self.defect_entry.parameters.update(gap_parameters)

    def run_compatibility(self):
        # Set potalign so pymatgen can calculate bandfilling for 'neutral' defects
        # (possible for resonant dopants etc.)
        if (
            self.defect_entry.charge == 0
            and "potalign" not in self.defect_entry.parameters
        ):
            self.defect_entry.parameters["potalign"] = 0

        self.defect_entry = self.compatibility.process_entry(self.defect_entry)

        if "delocalization_meta" in self.defect_entry.parameters:
            delocalization_meta = self.defect_entry.parameters["delocalization_meta"]
            if (
                "plnr_avg" in delocalization_meta
                and not delocalization_meta["plnr_avg"]["is_compatible"]
            ) or (
                "atomic_site" in delocalization_meta
                and not delocalization_meta["atomic_site"]["is_compatible"]
            ):
                specific_delocalized_warning = f"""
Delocalization analysis has indicated that {self.defect_entry.name}
with charge {self.defect_entry.charge} may not be compatible with the chosen charge correction."""
                general_delocalization_warning = """
Note: Defects throwing a "delocalization analysis" warning may require a larger supercell for
accurate total energies. Recommended to look at the correction plots (i.e. run 
`get_correction_freysoldt(DefectEntry,...,plot=True)` from
`doped.pycdt.corrections.finite_size_charge_correction`) to visually determine if the charge 
correction scheme is still appropriate (replace 'freysoldt' with 'kumagai' if using anisotropic 
correction). You can also change the DefectCompatibility() tolerance settings via the 
`compatibility` parameter in `SingleDefectParser.from_paths()`."""
                warnings.warn(message=specific_delocalized_warning)
                warnings.warn(
                    message=general_delocalization_warning
                )  # should only print once

        if "num_hole_vbm" in self.defect_entry.parameters:
            if (
                self.compatibility.free_chg_cutoff
                < self.defect_entry.parameters["num_hole_vbm"]
            ) or (
                self.compatibility.free_chg_cutoff
                < self.defect_entry.parameters["num_elec_cbm"]
            ):
                num_holes = self.defect_entry.parameters["num_hole_vbm"]
                num_electrons = self.defect_entry.parameters["num_elec_cbm"]
                warnings.warn(
                    f"Eigenvalue analysis has determined that `num_hole_vbm` (= {num_holes}) or "
                    f"`num_elec_cbm` (= {num_electrons}) is significant (>2.1) for "
                    f"{self.defect_entry.name} with charge {self.defect_entry.charge}, indicating "
                    f"that there are many free charges in this defect supercell calculation and "
                    f"so the defect charge correction is unlikely to be accurate."
                )
                if "freysoldt_meta" in self.defect_entry.parameters:
                    frey_meta = self.defect_entry.parameters["freysoldt_meta"]
                    frey_corr = (
                        frey_meta["freysoldt_electrostatic"]
                        + frey_meta["freysoldt_potential_alignment_correction"]
                    )
                    self.defect_entry.corrections.update(
                        {"charge_correction": frey_corr}
                    )
                elif "kumagai_meta" in self.defect_entry.parameters:
                    kumagai_meta = self.defect_entry.parameters["kumagai_meta"]
                    kumagai_corr = (
                        kumagai_meta["kumagai_electrostatic"]
                        + kumagai_meta["kumagai_potential_alignment_correction"]
                    )
                    self.defect_entry.corrections.update(
                        {"charge_correction": kumagai_corr}
                    )

        if (
            self.defect_entry.charge != 0
            and self.defect_entry.corrections.get("charge_correction", None) is None
        ):
            warnings.warn(
                f"No charge correction computed for {self.defect_entry.name} with "
                f"charge {self.defect_entry.charge}, indicating problems with the "
                f"required data for the charge correction (i.e. dielectric constant, "
                f"LOCPOT files for Freysoldt correction, OUTCAR (with ICORELEVEL = 0) "
                f"for Kumagai correction etc)."
            )


class PostProcess:
    def __init__(self, root_fldr, mpid=None, mapi_key=None):
        """
        Post processing object for charged point-defect calculations.

        Args:
            root_fldr (str): path (relative) to directory
                in which data of charged point-defect calculations for
                a particular system are to be found;
            mpid (str): Materials Project ID of bulk structure;
                format "mp-X", where X is an integer;
            mapi_key (str): Materials API key to access database.

        """
        self._root_fldr = root_fldr
        self._mpid = mpid
        self._mapi_key = mapi_key
        self._substitution_species = set()

    def parse_defect_calculations(self):
        """
        Parses the defect calculations as DefectEntry objects,
        from a PyCDT root_fldr file structure.
        Charge correction is missing in the first run.
        """
        logger = logging.getLogger(__name__)
        parsed_defects = []
        subfolders = glob.glob(os.path.join(self._root_fldr, "vac_*"))
        subfolders += glob.glob(os.path.join(self._root_fldr, "as_*"))
        subfolders += glob.glob(os.path.join(self._root_fldr, "sub_*"))
        subfolders += glob.glob(os.path.join(self._root_fldr, "inter_*"))

        def get_vr_and_check_locpot(fldr):
            vr_file = os.path.join(fldr, "vasprun.xml")
            if not (os.path.exists(vr_file) or os.path.exists(vr_file + ".gz")):
                logger.warning("{} doesn't exit".format(vr_file))
                error_msg = ": Failure, vasprun.xml doesn't exist."
                return (None, error_msg)  # Further processing is not useful

            try:
                vr, vr_path = get_vasprun(vr_file, parse_potcar_file=False)
            except:
                logger.warning("Couldn't parse {}".format(vr_file))
                error_msg = ": Failure, couldn't parse vasprun.xml file."
                return (None, error_msg)

            if not vr.converged:
                logger.warning("Vasp calculation at {} not converged".format(fldr))
                error_msg = ": Failure, Vasp calculation not converged."
                return (None, error_msg)  # Further processing is not useful

            # Check if locpot exists
            locpot_file = os.path.join(fldr, "LOCPOT")
            if not (os.path.exists(locpot_file) or os.path.exists(locpot_file + ".gz")):
                logger.warning("{} doesn't exit".format(locpot_file))
                error_msg = ": Failure, LOCPOT doesn't exist"
                return (None, error_msg)  # Further processing is not useful

            return (vr, None)

        def get_encut_from_potcar(fldr):
            potcar_file = os.path.join(fldr, "POTCAR")
            if not os.path.exists(potcar_file):
                logger.warning("Not POTCAR in {} to parse ENCUT".format(fldr))
                error_msg = ": Failure, No POTCAR file."
                return (None, error_msg)  # Further processing is not useful

            try:
                potcar = Potcar.from_file(potcar_file)
            except:
                logger.warning("Couldn't parse {}".format(potcar_file))
                error_msg = ": Failure, couldn't read POTCAR file."
                return (None, error_msg)

            encut = max(ptcr_sngl.enmax for ptcr_sngl in potcar)
            return (encut, None)

        # get bulk entry information first
        fldr = os.path.join(self._root_fldr, "bulk")
        vr, error_msg = get_vr_and_check_locpot(fldr)
        if error_msg:
            logger.error("Abandoning parsing of the calculations")
            return {}
        bulk_energy = vr.final_energy
        bulk_sc_struct = vr.final_structure
        try:
            encut = vr.incar["ENCUT"]
        except:  # ENCUT not specified in INCAR. Read from POTCAR
            encut, error_msg = get_encut_from_potcar(fldr)
            if error_msg:
                logger.error("Abandoning parsing of the calculations")
                return {}

        trans_dict = loadfn(os.path.join(fldr, "transformation.json"), cls=MontyDecoder)
        supercell_size = trans_dict["supercell"]

        bulk_file_path = fldr
        bulk_entry = ComputedStructureEntry(
            bulk_sc_struct,
            bulk_energy,
            data={
                "bulk_path": bulk_file_path,
                "encut": encut,
                "supercell_size": supercell_size,
            },
        )

        # get defect entry information
        for fldr in subfolders:
            fldr_name = os.path.split(fldr)[1]
            chrg_fldrs = glob.glob(os.path.join(fldr, "charge*"))
            for chrg_fldr in chrg_fldrs:
                trans_dict = loadfn(
                    os.path.join(chrg_fldr, "transformation.json"), cls=MontyDecoder
                )
                chrg = trans_dict["charge"]
                vr, error_msg = get_vr_and_check_locpot(chrg_fldr)
                if error_msg:
                    logger.warning("Parsing the rest of the calculations")
                    continue
                if (
                    "substitution_specie" in trans_dict
                    and trans_dict["substitution_specie"]
                    not in bulk_sc_struct.symbol_set
                ):
                    self._substitution_species.add(trans_dict["substitution_specie"])
                elif (
                    "inter" in trans_dict["defect_type"]
                    and trans_dict["defect_site"].specie.symbol
                    not in bulk_sc_struct.symbol_set
                ):
                    # added because extrinsic interstitials don't have
                    # "substitution_specie" character...
                    trans_dict["substitution_specie"] = trans_dict[
                        "defect_site"
                    ].specie.symbol
                    self._substitution_species.add(
                        trans_dict["defect_site"].specie.symbol
                    )

                defect_type = trans_dict.get("defect_type", None)
                energy = vr.final_energy
                try:
                    encut = vr.incar["ENCUT"]
                except:  # ENCUT not specified in INCAR. Read from POTCAR
                    encut, error_msg = get_encut_from_potcar(chrg_fldr)
                    if error_msg:
                        logger.warning(
                            "Not able to determine ENCUT " "in {}".format(fldr_name)
                        )
                        logger.warning("Parsing the rest of the " "calculations")
                        continue

                comp_data = {
                    "bulk_path": bulk_file_path,
                    "defect_path": chrg_fldr,
                    "encut": encut,
                    "fldr_name": fldr_name,
                    "supercell_size": supercell_size,
                }
                if "substitution_specie" in trans_dict:
                    comp_data["substitution_specie"] = trans_dict["substitution_specie"]

                # create Defect Object as dict, then load to DefectEntry object
                defect_dict = {
                    "structure": bulk_sc_struct,
                    "charge": chrg,
                    "@module": "pymatgen.analysis.defects.core",
                }
                defect_site = trans_dict["defect_supercell_site"]
                if "vac_" in defect_type:
                    defect_dict["@class"] = "Vacancy"
                elif "as_" in defect_type or "sub_" in defect_type:
                    defect_dict["@class"] = "Substitution"
                    substitution_specie = trans_dict["substitution_specie"]
                    defect_site = PeriodicSite(
                        substitution_specie,
                        defect_site.frac_coords,
                        defect_site.lattice,
                        coords_are_cartesian=False,
                    )
                elif "int_" in defect_type:
                    defect_dict["@class"] = "Interstitial"
                else:
                    raise ValueError(
                        "defect type {} not recognized...".format(defect_type)
                    )

                defect_dict.update({"defect_site": defect_site})
                defect = MontyDecoder().process_decoded(defect_dict)
                parsed_defects.append(
                    DefectEntry(defect, energy - bulk_energy, parameters=comp_data)
                )

        try:
            parsed_defects_data = {}
            parsed_defects_data["bulk_entry"] = bulk_entry
            parsed_defects_data["defects"] = parsed_defects
            return parsed_defects_data
        except:
            return {}  # Return Null dict due to failure

    def get_vbm_bandgap(self):
        """
        Returns the valence band maximum (float) of the structure with
        MP-ID mpid.

        Args:
            mpid (str): MP-ID for which the valence band maximum is to
                be fetched from the Materials Project database
        """
        logger = logging.getLogger(__name__)
        vbm, bandgap = None, None

        if self._mpid is not None:
            with MPRester(api_key=self._mapi_key) as mp:
                bs = mp.get_bandstructure_by_material_id(self._mpid)
            if bs:
                vbm = bs.get_vbm()["energy"]
                bandgap = bs.get_band_gap()["energy"]

        if vbm is None or bandgap is None:
            if self._mpid:
                logger.warning(
                    "Mpid {} was provided, but no bandstructure entry currently exists for it. "
                    "Reverting to use of bulk calculation.".format(self._mpid)
                )
            else:
                logger.warning(
                    "No mp-id provided, will fetch CBM/VBM details from the "
                    "bulk calculation."
                )
            logger.warning(
                "This may not be appropriate if the VBM/CBM occur at reciprocal points "
                "not included in the bulk calculation."
            )
            vr, vr_path = get_vasprun(
                os.path.join(self._root_fldr, "bulk", "vasprun.xml"),
                parse_potcar_file=False,
            )
            bandgap = vr.eigenvalue_band_properties[0]
            vbm = vr.eigenvalue_band_properties[2]

        return (vbm, bandgap)

    def get_chempot_limits(self):
        """
        Returns atomic chempots from bulk_composition based on data in
        the materials project database. This is abstractly handled in the
        ChemPotAnalyzer

        Note to user: If personal phase diagram desired,
            option exists in the pycdt.core.chemical_potentials to setup,
            run and parse personal phase diagrams for purposes of chemical potentials
        """
        logger = logging.getLogger(__name__)

        if self._mpid:
            cpa = chemical_potentials.MPChemPotAnalyzer(
                mpid=self._mpid,
                sub_species=self._substitution_species,
                mapi_key=self._mapi_key,
            )
        else:
            bulkvr, bulkvr_path = get_vasprun(
                os.path.join(self._root_fldr, "bulk", "vasprun.xml"),
                parse_potcar_file=False,
            )
            if not bulkvr:
                msg = "Could not fetch computed entry for atomic chempots!"
                logger.warning(msg)
                raise ValueError(msg)
            cpa = chemical_potentials.MPChemPotAnalyzer(
                bulk_ce=bulkvr.get_computed_entry(),
                sub_species=self._substitution_species,
                mapi_key=self._mapi_key,
            )

        chem_lims = cpa.analyze_GGA_chempots()

        return chem_lims

    def parse_dielectric_calculation(self):
        """
        Parses the "vasprun.xml(.gz)" file in subdirectory "dielectric" of
        root directory root_fldr and returns the average of the trace
        of the dielectric tensor.

        Args:
            root_fldr (str):
                root directory where subdirectory "dielectric" is expected
        Returns:
            eps (float):
                average of the trace of the dielectric tensor
        """

        try:
            vr, vr_path = get_vasprun(
                os.path.join(self._root_fldr, "dielectric", "vasprun.xml"),
                parse_potcar_file=False,
            )
        except:
            logging.getLogger(__name__).warning("Parsing Dielectric calculation failed")
            return None

        eps_ion = vr.epsilon_ionic
        eps_stat = vr.epsilon_static

        eps = []
        for i in range(len(eps_ion)):
            eps.append([e[0] + e[1] for e in zip(eps_ion[i], eps_stat[i])])

        return eps

    def compile_all(self):
        """
        Run to get all post processing objects as dictionary

        note: still need to implement
            1) ability for substitutional atomic chempots
            2) incorporated charge corrections for defects
        """
        output = self.parse_defect_calculations()
        output["epsilon"] = self.parse_dielectric_calculation()
        output["mu_range"] = self.get_chempot_limits()
        vbm, gap = self.get_vbm_bandgap()
        output["vbm"] = vbm
        output["gap"] = gap

        return output


def get_site_mapping_indices(
    structure_a: Structure, structure_b: Structure, threshold=2.0
):
    """
    Reset the position of a partially relaxed structure to its unrelaxed positions.
    The template structure may have a different species ordering to the `input_structure`.
    """

    ## Generate a site matching table between the input and the template
    input_fcoords = [site.frac_coords for site in structure_a]
    template_fcoords = [site.frac_coords for site in structure_b]

    dmat = structure_a.lattice.get_all_distances(input_fcoords, template_fcoords)
    min_dist_with_index = []
    for index in range(len(input_fcoords)):
        dists = dmat[index]
        template_index = dists.argmin()
        current_dist = dists.min()
        min_dist_with_index.append(
            [
                current_dist,
                index,
                template_index,
            ]
        )

        if current_dist > threshold:
            sitea = structure_a[index]
            siteb = structure_b[template_index]
            warnings.warn(
                f"Large site displacement {current_dist:.4f} detected when matching atomic sites:"
                f" {sitea}-> {siteb}."
            )
    return min_dist_with_index


def reorder_unrelaxed_structure(
    unrelaxed_structure: Structure, initial_relax_structure: Structure, threshold=2.0
):
    """
    Reset the position of a partially relaxed structure to its unrelaxed positions.
    The template structure may have a different species ordering to the `input_structure`.
    """

    # Obtain site mapping between the initial_relax_structure and the unrelaxed structure
    mapping = get_site_mapping_indices(
        initial_relax_structure, unrelaxed_structure, threshold=threshold
    )

    # Reorder the unrelaxed_structure so it matches the ordering of the initial_relax_structure (
    # from the actual calculation)
    reordered_sites = [unrelaxed_structure[tmp[2]] for tmp in mapping]
    new_structure = Structure.from_sites(reordered_sites)

    assert len(new_structure) == len(unrelaxed_structure)

    return new_structure



# coding: utf-8

from __future__ import division

__status__ = "Development"

import os

from pymatgen.core import PeriodicSite
from pymatgen.core.structure import Structure
from pymatgen.util.testing import PymatgenTest

from doped.pycdt.core.defectsmaker import *

file_loc = os.path.abspath(os.path.join(__file__, "..", "..", "..", "..", "test_files"))


class GetOptimizedScScaleTest(PymatgenTest):
    def setUp(self):
        self.gaas_prim_struct = Structure.from_file(
            os.path.join(file_loc, "POSCAR_GaAs")
        )

    def test_biggercell_wanted(self):
        lattchange = get_optimized_sc_scale(self.gaas_prim_struct, 300)
        self.assertEqual([5, 5, 5], lattchange)
        lattchange = get_optimized_sc_scale(self.gaas_prim_struct, 100)
        self.assertEqual([3, 3, 3], lattchange)


class DefectChargerSemiconductorTest(PymatgenTest):
    def setUp(self):
        self.gaas_struct = Structure.from_file(os.path.join(file_loc, "POSCAR_GaAs"))
        self.def_charger = DefectChargerSemiconductor(self.gaas_struct)

    def test_vacancy_charges(self):
        """
        Reference: PRB 71, 125207 (2005)
        """
        ga_vac_qs = self.def_charger.get_charges("vacancy", "Ga")
        as_vac_qs = self.def_charger.get_charges("vacancy", "As")
        self.assertIn(0, ga_vac_qs)
        self.assertIn(-3, ga_vac_qs)
        self.assertIn(1, as_vac_qs)
        self.assertIn(-3, as_vac_qs)

    def test_antisite_charges(self):
        ga_on_as_qs = self.def_charger.get_charges("antisite", "Ga", "As")
        as_on_ga_qs = self.def_charger.get_charges("antisite", "As", "Ga")
        self.assertIn(0, ga_on_as_qs)
        self.assertIn(2, ga_on_as_qs)
        self.assertIn(0, as_on_ga_qs)
        self.assertIn(-3, as_on_ga_qs)

    def test_substitution_charges(self):
        s_impurity_qs = self.def_charger.get_charges("substitution", "As", "S")
        se_impurity_qs = self.def_charger.get_charges("substitution", "As", "Se")
        mg_impurity_qs = self.def_charger.get_charges("substitution", "Ga", "Mg")
        zn_impurity_qs = self.def_charger.get_charges("substitution", "Ga", "Zn")
        self.assertArrayEqual(s_impurity_qs, [-1, 0, 1, 2, 3, 4, 5, 6])
        self.assertArrayEqual(se_impurity_qs, [-1, 0, 1, 2, 3, 4, 5, 6])
        self.assertArrayEqual(mg_impurity_qs, [-2, -1, 0, 1, 2])
        self.assertArrayEqual(zn_impurity_qs, [-2, -1, 0, 1, 2])

    def test_interstitial_charges(self):
        """
        References:
        N interstitial: +1 to -3 [J. Phys.: Condens. Matter 20 (2008) 235231]
        As Self interstital: arxiv.org/pdf/1101.1413.pdf
        """
        n_qs = self.def_charger.get_charges("interstitial", "N")
        self.assertIn(1, n_qs)
        self.assertIn(-3, n_qs)
        self_qs = self.def_charger.get_charges("interstitial", "As")
        self.assertIn(-1, self_qs)
        self.assertIn(1, self_qs)


class DefectChargerInsulatorTest(PymatgenTest):
    def setUp(self):
        cr2o3_struct = Structure.from_file(os.path.join(file_loc, "POSCAR_Cr2O3"))
        self.def_charger = DefectChargerInsulator(cr2o3_struct)

    def test_vacancy_charges(self):
        """
        For insulators, the range of defect charges expected is
        -A to 0 or 0 to B, where A is cation charge in its native
        oxidation state in the compound, and B is the corresponding
        anion charge.
        """
        cr_vac_qs = self.def_charger.get_charges("vacancy", "Cr")
        o_vac_qs = self.def_charger.get_charges("vacancy", "O")
        self.assertIn(0, cr_vac_qs)
        self.assertIn(-3, cr_vac_qs)
        self.assertNotIn(-4, cr_vac_qs)
        self.assertNotIn(1, cr_vac_qs)
        self.assertIn(0, o_vac_qs)
        self.assertIn(2, o_vac_qs)
        self.assertNotIn(3, o_vac_qs)
        self.assertNotIn(-1, o_vac_qs)

    def test_antisite_charges(self):
        """
        Anitisites are not expected for insulators.
        Skipping this.
        """
        pass

    def test_substitution_charges(self):
        ti_on_cr_qs = self.def_charger.get_charges("substitution", "Cr", "Ti")
        self.assertIn(0, ti_on_cr_qs)
        self.assertIn(1, ti_on_cr_qs)
        self.assertNotIn(-1, ti_on_cr_qs)
        self.assertNotIn(2, ti_on_cr_qs)
        mg_on_cr_qs = self.def_charger.get_charges("substitution", "Cr", "Mg")
        self.assertIn(-1, mg_on_cr_qs)
        self.assertNotIn(0, mg_on_cr_qs)
        self.assertNotIn(-2, mg_on_cr_qs)
        self.assertNotIn(1, mg_on_cr_qs)

    def test_interstitial_charges(self):
        ti_inter_qs = self.def_charger.get_charges("interstitial", "Ti")
        self.assertIn(0, ti_inter_qs)
        self.assertIn(4, ti_inter_qs)
        self.assertNotIn(-1, ti_inter_qs)
        self.assertNotIn(5, ti_inter_qs)


class ChargedDefectsStructuresTest(PymatgenTest):
    def setUp(self):
        self.gaas_struct = Structure.from_file(os.path.join(file_loc, "POSCAR_GaAs"))
        self.ga_site = self.gaas_struct.sites[0]
        self.as_site = self.gaas_struct.sites[1]

    def test_simple_initialization(self):
        # test simple attributes
        CDS = ChargedDefectsStructures(self.gaas_struct)
        self.assertIsInstance(CDS.struct, Structure)
        self.assertIsInstance(CDS.defect_charger, DefectChargerSemiconductor)
        self.assertFalse(CDS.substitutions)
        superstruct = CDS.defects["bulk"]["supercell"]["structure"]
        self.assertIsInstance(superstruct, Structure)
        cellsize = CDS.defects["bulk"]["supercell"]["size"]
        self.assertEqual([4, 4, 4], cellsize)

        # test native (intrinsic) defect generation
        self.assertEqual("vac_1_Ga", CDS.defects["vacancies"][0]["name"])
        self.assertEqual(self.ga_site, CDS.defects["vacancies"][0]["unique_site"])
        self.assertEqual("vac_2_As", CDS.defects["vacancies"][1]["name"])
        self.assertEqual(self.as_site, CDS.defects["vacancies"][1]["unique_site"])

        self.assertEqual("as_1_As_on_Ga", CDS.defects["substitutions"][0]["name"])
        self.assertEqual(self.ga_site, CDS.defects["substitutions"][0]["unique_site"])
        self.assertEqual("as_1_Ga_on_As", CDS.defects["substitutions"][1]["name"])
        self.assertEqual(self.as_site, CDS.defects["substitutions"][1]["unique_site"])

    def test_extra_initialization(self):
        CDS = ChargedDefectsStructures(
            self.gaas_struct, cellmax=513, struct_type="insulator", antisites_flag=False
        )
        self.assertIsInstance(CDS.defect_charger, DefectChargerInsulator)
        cellsize = CDS.defects["bulk"]["supercell"]["size"]
        self.assertEqual([5, 5, 5], cellsize)
        self.assertFalse(len(CDS.defects["substitutions"]))  # testing antisite flag

    def test_subs_and_interstits(self):
        # test manual subtitution specification
        CDS = ChargedDefectsStructures(
            self.gaas_struct,
            antisites_flag=False,
            substitutions={"Ga": ["Si", "In"], "As": ["Sb"]},
        )
        check_subs = {
            sub["name"]: sub["unique_site"] for sub in CDS.defects["substitutions"]
        }
        self.assertEqual(3, len(check_subs))
        self.assertEqual(self.ga_site, check_subs["sub_1_Si_on_Ga"])
        self.assertEqual(self.ga_site, check_subs["sub_1_In_on_Ga"])
        self.assertEqual(self.as_site, check_subs["sub_2_Sb_on_As"])

        # Test automatic interstitial finding.
        CDS = ChargedDefectsStructures(
            self.gaas_struct, include_interstitials=True, interstitial_elements=["Mn"]
        )
        self.assertEqual(CDS.get_n_defects_of_type("interstitials"), 2)
        fnames = [
            i["name"][i["name"].index("M") :] for i in CDS.defects["interstitials"]
        ]
        self.assertEqual(sorted(fnames), sorted(["Mn_InFiT1_mult6", "Mn_InFiT2_mult6"]))
        nsites = len(CDS.defects["interstitials"][0]["supercell"]["structure"].sites)
        self.assertEqual(
            len(CDS.get_ith_supercell_of_defect_type(0, "interstitials").sites), nsites
        )
        self.assertEqual(
            CDS.defects["interstitials"][0]["charges"], [0, 1, 2, 3, 4, 5, 6, 7]
        )
        self.assertEqual(
            CDS.defects["interstitials"][1]["charges"], [0, 1, 2, 3, 4, 5, 6, 7]
        )

        # Test manual interstitial specification.
        isite = PeriodicSite(
            "Mn",
            CDS.defects["interstitials"][0]["supercell"]["structure"][
                nsites - 1
            ].coords,
            self.gaas_struct.lattice,
        )
        cds2 = ChargedDefectsStructures(
            self.gaas_struct,
            include_interstitials=True,
            standardized=False,
            # standardized = False is required when manually specifying interstitial,
            # because dont want base structure lattice to change
            intersites=(isite,),
        )
        self.assertEqual(cds2.get_n_defects_of_type("interstitials"), 2)
        fnames = [
            i["name"][i["name"].index("_") + 3 :] for i in cds2.defects["interstitials"]
        ]
        self.assertEqual(sorted(fnames), sorted(["As", "Ga"]))
        nsites = len(cds2.defects["interstitials"][0]["supercell"]["structure"].sites)
        self.assertEqual(
            len(cds2.get_ith_supercell_of_defect_type(0, "interstitials").sites), nsites
        )
        cds3 = ChargedDefectsStructures(
            self.gaas_struct,
            include_interstitials=True,
            standardized=False,
            # standardized = False is required when manually specifying interstitial,
            # because dont want base structure lattice to change
            interstitial_elements=["Mn"],
            intersites=(isite,),
        )
        self.assertEqual(cds3.get_n_defects_of_type("interstitials"), 1)
        fnames = [
            i["name"][i["name"].index("_") + 3 :] for i in cds3.defects["interstitials"]
        ]
        self.assertEqual(sorted(fnames), sorted(["Mn"]))
        nsites = len(cds3.defects["interstitials"][0]["supercell"]["structure"].sites)
        self.assertEqual(
            len(cds3.get_ith_supercell_of_defect_type(0, "interstitials").sites), nsites
        )


if __name__ == "__main__":
    import unittest

    unittest.main()



# coding: utf-8

from __future__ import division

__status__ = "Development"

import unittest

from doped.pycdt.utils.units import eV_to_k, k_to_eV


class EnergyFunctionsTest(unittest.TestCase):
    def test_k_to_eV(self):
        g = [0.1, 0.2, 0.3]
        self.assertAlmostEqual(k_to_eV(g), 0.5333804)

    def test_eV_to_k(self):
        self.assertAlmostEqual(eV_to_k(1.0), 0.9681404248678961)


import unittest

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

__status__ = "Development"

import logging
import logging.config
import os

from monty.serialization import loadfn

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))


def initialize_logging(filename=None, level=None):
    config_dict = loadfn(os.path.join(MODULE_DIR, "logging.yaml"))
    if filename:
        config_dict["handlers"]["file_handler"].update({"filename": filename})
    if level:
        config_dict["handlers"]["file_handler"].update({"level": level})

    logging.config.dictConfig(config_dict)



"""
This module contains unit conversion constants and  functions
"""

import math
import warnings

import numpy as np

# Define conversion_constants
hart_to_ev = 27.2114
ang_to_bohr = 1.8897
invang_to_ev = 3.80986
kb = 8.6173324e-5  # eV / K
hbar = 6.58211928e-16  # eV s

conv = (
    math.sqrt((9.1 * 1e-31) ** 3)
    * math.sqrt((1.6 * 1e-19) ** 3)
    / ((1.05 * 1e-34) ** 3)
)

warnings.warn(
    "Replaced PyCDT usage of utils objects with "
    "objects from pymatgen.analysis.defects\n"
    "Will remove k_to_eV with Version 2.5 of PyCDT.",
    DeprecationWarning,
)


def k_to_eV(g):
    """
    Convert a k-vector to energy [eV] via hbar*k^2/2m
    Args:
        a: Reciprocal vector (units of 1/A).

    Returns:
        (double) Energy in eV
    """
    return invang_to_ev * np.dot(g, g)


warnings.warn(
    "Replaced PyCDT usage of utils objects with "
    "objects from pymatgen.analysis.defects\n"
    "Will remove eV_to_k with Version 2.5 of PyCDT.",
    DeprecationWarning,
)


def eV_to_k(energy):
    """
    Convert energy to reciprocal vector magnitude k via hbar*k^2/2m
    Args:
        a: Energy in eV.

    Returns:
        (double) Reciprocal vector magnitude (units of 1/Bohr).
    """
    return math.sqrt(energy / invang_to_ev) * ang_to_bohr



# coding: utf-8

from __future__ import division

__status__ = "Development"

import os

from pymatgen.analysis.defects.core import DefectEntry, Vacancy
from pymatgen.analysis.defects.thermodynamics import DefectPhaseDiagram
from pymatgen.core import Element
from pymatgen.core.structure import Lattice, PeriodicSite, Structure

# from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.util.testing import PymatgenTest

from doped.pycdt.utils.plotter import DefectPlotter


class DefectPlotterTest(PymatgenTest):
    def setUp(self):
        l = Lattice([[3.52, 0.0, 2.033], [1.174, 3.32, 2.033], [0.0, 0.0, 4.066]])
        s_bulk = Structure(
            l, ["Ga", "As"], [[0.0000, 0.0000, 0.0000], [0.2500, 0.2500, 0.2500]]
        )
        defect_site = PeriodicSite("As", [0.25, 0.25, 0.25], l)
        defect = Vacancy(s_bulk, defect_site, charge=1.0)
        defect_entry = DefectEntry(defect, 0.0)

        entries = [defect_entry]
        vbm = 0.2
        band_gap = 1.0
        dpd = DefectPhaseDiagram(entries, vbm, band_gap)
        self.dp = DefectPlotter(dpd)

    def test_get_plot_form_energy(self):
        mu_elts = {Element("As"): 0, Element("Ga"): 0}
        self.dp.get_plot_form_energy(mu_elts).savefig("test.pdf")
        self.assertTrue(os.path.exists("test.pdf"))
        os.system("rm test.pdf")

    # def test_plot_conc_temp(self):
    #     self.dp.plot_conc_temp().savefig('test.pdf')
    #     self.assertTrue(os.path.exists('test.pdf'))
    #     os.system('rm test.pdf')
    #
    # def test_plot_carriers_ef(self):
    #     self.dp.plot_carriers_ef().savefig('test.pdf')
    #     self.assertTrue(os.path.exists('test.pdf'))
    #     os.system('rm test.pdf')

    # def tearDown(self):
    #     self.da


import unittest

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""
This is just for plotting the axis files that come out of the sxdefectalignwrapper script. 
Created 8/11/15 (right after I finished the first version of the wrapper class)
"""

__status__ = "Development"

import os
import warnings

import matplotlib.pyplot as plt
import numpy as np
from pymatgen.io.vasp.outputs import Locpot

warnings.warn(
    "Replacing PyCDT usage of Freysoldt base classes with calls to "
    "corresponding objects in pymatgen.analysis.defects.corrections\n"
    "All correction plotting functionalities exist within pymatgen v2019.5.1."
    "Version 2.5 of PyCDT will remove pycdt.corrections.freysoldt_plotter.FreysoldtPlot.",
    DeprecationWarning,
)


class FreysoldtPlot(object):
    """
        NOTE from developers:
            This code has excessive electrostatic potential
            plotting methods. This code will not ever be used
            by command line option and has unit test
            It will not be maintained
            going forward (as of 12/15/2017).
            However, we are keeping function here to allow for
            current users to make use of it...
    This class applies the Freysoldt correction to remove electrostatic defect
    interaction contribution to energy and apply potential alignment.
    Ideally this sxdefectalign wrapper class should be replaced by a python
    code implementing a generalized anisotropic Freysoldt correction
    """

    def __init__(
        self, pathforaxis, site_frac_coords, name="", locpotbulk="", locpotdef=""
    ):
        """
        Args:
            pathforaxis:
                path for axis1vline-ev.dat type files.
            site_frac_coords:
                Fractional coordinates of defect site as list
            name:
                Name of the defect to write files
        """
        self._path = pathforaxis
        self._frac_coords = site_frac_coords
        self._name = name
        self._locpot_bulk = locpotbulk
        self._locpot_defect = locpotdef

    def plot_hartree_pot(self):
        # plot planar averages of bulk and defect (good for seeing global changes)
        if not self._locpot_bulk:
            print(
                "did not feed path to locpot file! Need this for plotting hartree pot"
            )
            return
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(3, 1, 1)
        ax.set_title("Locpot planar averaged potentials")
        bulkloc = Locpot.from_file(self._locpot_bulk)
        defloc = Locpot.from_file(self._locpot_defect)
        get_agrid = bulkloc.get_axis_grid
        get_baavg = bulkloc.get_average_along_axis
        get_daavg = defloc.get_average_along_axis
        for axis in [0, 1, 2]:
            ax = fig.add_subplot(3, 1, axis + 1)
            latt_len = self._lengths[axis]
            ax.plot(get_agrid(axis), get_baavg(axis), "r", label="Bulk potential")
            ax.plot(get_agrid(axis), get_daavg(axis), "b", label="Defect potential")
            ax.plot(
                [self._frac_coords[axis] * latt_len],
                [0],
                "or",
                markersize=4.0,
                label="Defect site",
            )
            ax.set_ylabel("axis " + str(axis + 1))
            if not axis:
                ax.legend()
        ax.set_xlabel("distance (Angstrom)")
        # plt.savefig('locpotavgplot.png')
        plt.show()

    def plot_hartree_pot_diff(self):
        # only plot the difference in planar averaged potentials
        if not self._locpot_bulk:
            print(
                "did not feed path to locpot file! Need this for plotting hartree pot diff"
            )
            return
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(3, 1, 1)
        ax.set_title("Locpot planar averaged potential difference")
        bulkloc = Locpot.from_file(self._locpot_bulk)
        defloc = Locpot.from_file(self._locpot_defect)
        for axis in [0, 1, 2]:
            ax = fig.add_subplot(3, 1, axis + 1)
            defect_axis = defloc.get_axis_grid(axis)
            defect_pot = defloc.get_average_along_axis(axis)
            pure_pot = bulkloc.get_average_along_axis(axis)
            latt_len = self._lengths[axis]
            ax.plot(
                defect_axis, defect_pot - pure_pot, "b", label="Defect-Bulk difference"
            )
            ax.plot(
                [self._frac_coords[axis] * latt_len],
                [0],
                "or",
                markersize=4.0,
                label="Defect site",
            )
            ax.set_ylabel("axis " + str(axis + 1))
            if not axis:
                ax.legend()
        ax.set_xlabel("distance (Angstrom)")
        # plt.savefig('locpotdiffplot.png')
        plt.show()

    def plot_all_hartree_pot(self):
        # plot planar averaged locpots, along with the difference for each axis
        if not self._locpot_bulk:
            print(
                "did not feed path to locpot file! Need this for plotting hartree pot"
            )
            return
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(3, 1, 1)
        ax.set_title("Locpot planar averaged potentials and difference")
        bulkloc = Locpot.from_file(self._locpot_bulk)
        defloc = Locpot.from_file(self._locpot_defect)
        for axis in [0, 1, 2]:
            ax = fig.add_subplot(3, 1, axis + 1)
            defect_axis = defloc.get_axis_grid(axis)
            defect_pot = defloc.get_average_along_axis(axis)
            pure_pot = bulkloc.get_average_along_axis(axis)
            ax.plot(defect_axis, pure_pot, "r", label="Bulk potential")
            ax.plot(defect_axis, defect_pot, "b", label="Defect potential")
            ax.plot(
                defect_axis, defect_pot - pure_pot, "k", label="Defect-Bulk difference"
            )
            ax.plot(
                [self._frac_coords[axis] * self._lengths[axis]],
                [0],
                "or",
                markersize=4.0,
                label="Defect site",
            )
            ax.set_ylabel("axis " + str(axis + 1))
            if not axis:
                ax.legend()
        ax.set_xlabel("distance (Angstrom)")
        # plt.savefig('locpotavgdiffplot.png')
        plt.show()

    def plot_vline(self):
        """
        Will plot the vline files based on whatever local directory you provided
        """
        plotvals = {"0": {}, "1": {}, "2": {}}
        platy = []
        for axis in [0, 1, 2]:
            print("do axis " + str(axis + 1))
            x_lr, y_lr = [], []
            x, y = [], []
            x_diff, y_diff = [], []
            loc = os.path.abspath(self._path)
            if not self._name:
                nom = str(loc) + "/axis" + str(axis) + "vline-eV.dat"
            else:
                nom = (
                    str(loc)
                    + "/"
                    + str(self._name)
                    + "axis"
                    + str(axis)
                    + "vline-eV.dat"
                )
            with open(nom, "r") as f_sr:  # read in potential
                for r in f_sr:
                    tmp = r.split("\t")
                    if len(tmp) < 3 and not r.startswith("&"):
                        x_lr.append(float(tmp[0]) / 1.889725989)
                        y_lr.append(float(tmp[1]))
                    if len(tmp) > 2:
                        x.append(float(tmp[0]) / 1.889725989)
                        x_diff.append(float(tmp[0]) / 1.889725989)
                        y.append(float(tmp[2].rstrip("\n")))
                        y_diff.append(float(tmp[1]))

            # Extract potential alignment term averaging window of +/- 1 Ang
            # around point halfway between neighboring defects

            latt_len = max(x_diff)
            if self._frac_coords[axis] >= 0.5:
                platx = (self._frac_coords[axis] - 0.5) * latt_len
            else:
                platx = (self._frac_coords[axis] + 0.5) * latt_len
            print("half way between defects is: ", platx)

            xmin = latt_len - (1 - platx) if platx < 1 else platx - 1
            xmax = 1 - (latt_len - platx) if platx > latt_len - 1 else 1 + platx
            print("means sampling region is (", xmin, ",", xmax, ")")

            tmpalign = []
            if xmax < xmin:
                print("wrap around detected, special alignment needed")
                for i in range(len(x)):
                    if x[i] < xmax or x[i] > xmin:
                        tmpalign.append(y[i])
                    else:
                        continue
            else:
                for i in range(len(x)):
                    if x[i] > xmin and x[i] < xmax:
                        tmpalign.append(y[i])
                    else:
                        continue

            print("alignment is ", -np.mean(tmpalign))
            platy.append(-np.mean(tmpalign))
            flag = 0
            for i in tmpalign:  # check to see if alignment region varies too much
                if np.abs(i - platy[-1]) > 0.2:
                    flag = 1
                else:
                    continue
            if flag != 0:
                print(
                    "Warning: potential aligned region varied by more "
                    + "than 0.2eV (in range of halfway between defects "
                    + "+/-1 \Angstrom). Might have issues with Freidel "
                    + "oscillations or atomic relaxation\n"
                )

            plotvals[str(axis)]["xylong"] = [x_lr, y_lr]
            plotvals[str(axis)]["xy"] = [x, y]
            plotvals[str(axis)]["xydiff"] = [x_diff, y_diff]

        fig = plt.figure(figsize=(15.0, 12.0))
        for axis in [0, 1, 2]:
            print("plot axis ", axis + 1)
            ax = fig.add_subplot(3, 1, axis + 1)
            ax.set_ylabel("axis " + str(axis + 1))
            vals_plot = plotvals[str(axis)]
            ax.plot(vals_plot["xy"][0], vals_plot["xy"][1])
            ax.plot(vals_plot["xydiff"][0], vals_plot["xydiff"][1], "r")
            ax.plot(vals_plot["xylong"][0], vals_plot["xylong"][1], "g")

            if not axis:
                ax.set_title("Electrostatic planar averaged potential")
                ax.legend(["V_defect-V_ref-V_lr", "V_defect-V_ref", "V_lr"])

            latt_len = max(vals_plot["xydiff"][0])
            fcoords = self._frac_coords[axis]
            ax.plot([fcoords * latt_len], [0], "or", markersize=4.0)
            if fcoords >= 0.5:
                ax.plot([(fcoords - 0.5) * latt_len], [0], "og", markersize=4.0)
            else:
                ax.plot([(fcoords + 0.5) * latt_len], [0], "og", markersize=4.0)
            plt.axhline(y=0.0, linewidth=0.8, color="black")

        if not self._name:
            plt.savefig(os.path.join(self._path, "locpotgraph.png"))
        else:
            plt.savefig(os.path.join(self._path, str(self._name) + "locpotgraph.png"))

        return


if __name__ == "__main__":
    pass



import itertools
import logging
from math import erfc, exp

import numpy as np

# These functions are taken from the AIDE package developed by
# Adam Jackson and Alex Ganose (https://github.com/SMTG-UCL/aide)


def get_image_charge_correction(
    lattice,
    dielectric_matrix,
    conv=0.3,
    factor=30,
    motif=[0.0, 0.0, 0.0],
    verbose=False,
):
    """Calculates the anisotropic image charge correction by Sam Murphy in eV.

    This a rewrite of the code 'madelung.pl' written by Sam Murphy (see [1]).
    The default convergence parameter of conv = 0.3 seems to work perfectly
    well. However, it may be worth testing convergence of defect energies with
    respect to the factor (i.e. cut-off radius).

    References:
        [1] S. T. Murphy and N. D. H. Hine, Phys. Rev. B 87, 094111 (2013).

    Args:
        lattice (list): The defect cell lattice as a 3x3 matrix.
        dielectric_matrix (list): The dielectric tensor as 3x3 matrix.
        conv (float): A value between 0.1 and 0.9 which adjusts how much real
                      space vs reciprocal space contribution there is.
        factor: The cut-off radius, defined as a multiple of the longest cell
            parameter.
        motif: The defect motif (doesn't matter for single point defects, but
            included in case we include the extended code for defect clusters).
        verbose (bool): If True details of the correction will be printed.

    Returns:

        The image charge correction as {charge: correction}
    """
    inv_diel = np.linalg.inv(dielectric_matrix)
    det_diel = np.linalg.det(dielectric_matrix)
    latt = np.sqrt(np.sum(lattice**2, axis=1))

    # calc real space cutoff
    longest = max(latt)
    r_c = factor * longest

    # Estimate the number of boxes required in each direction to ensure
    # r_c is contained (the tens are added to ensure the number of cells
    # contains r_c). This defines the size of the supercell in which
    # the real space section is performed, however only atoms within rc
    # will be conunted.
    axis = np.array([int(r_c / a + 10) for a in latt])

    # Calculate supercell parallelpiped and dimensions
    sup_latt = np.dot(np.diag(axis), lattice)

    # Determine which of the lattice parameters is the largest and determine
    # reciprocal space supercell
    recip_axis = np.array([int(x) for x in factor * max(latt) / latt])
    recip_volume = abs(np.dot(np.cross(lattice[0], lattice[1]), lattice[2]))

    # Calculatate the reciprocal lattice vectors (need factor of 2 pi)
    recip_latt = np.linalg.inv(lattice).T * 2 * np.pi

    real_space = _get_real_space(
        conv, inv_diel, det_diel, latt, longest, r_c, axis, sup_latt
    )
    reciprocal = _get_recip(
        conv,
        inv_diel,
        det_diel,
        latt,
        recip_axis,
        recip_volume,
        recip_latt,
        dielectric_matrix,
    )

    # calculate the other terms and the final Madelung potential
    third_term = -2 * conv / np.sqrt(np.pi * det_diel)
    fourth_term = -3.141592654 / (recip_volume * conv**2)
    madelung = -(real_space + reciprocal + third_term + fourth_term)

    # convert to atomic units
    conversion = 14.39942
    real_ev = real_space * conversion / 2
    recip_ev = reciprocal * conversion / 2
    third_ev = third_term * conversion / 2
    fourth_ev = fourth_term * conversion / 2
    madelung_ev = madelung * conversion / 2

    correction = {}
    for q in range(1, 8):
        makov = 0.5 * madelung * q**2 * conversion
        lany = 0.65 * makov
        correction[q] = makov

    # TODO: Use tabulate
    if verbose:
        logging.info(
            """
    Results                      v_M^scr    dE(q=1) /eV
    -----------------------------------------------------
    Real space contribution    =  {:.6f}     {:.6f}
    Reciprocal space component =  {:.6f}     {:.6f}
    Third term                 = {:.6f}    {:.6f}
    Neutralising background    = {:.6f}    {:.6f}
    -----------------------------------------------------
    Final Madelung potential   = {:.6f}     {:.6f}
    -----------------------------------------------------""".format(
                real_space,
                real_ev,
                reciprocal,
                recip_ev,
                third_term,
                third_ev,
                fourth_term,
                fourth_ev,
                madelung,
                madelung_ev,
            )
        )

        logging.info(
            """
    Here are your final corrections:
    +--------+------------------+-----------------+
    | Charge | Point charge /eV | Lany-Zunger /eV |
    +--------+------------------+-----------------+"""
        )
        for q in range(1, 8):
            makov = 0.5 * madelung * q**2 * conversion
            lany = 0.65 * makov
            correction[q] = makov
            logging.info(
                "|   {}    |     {:10f}   |    {:10f}   |".format(q, makov, lany)
            )
        logging.info("+--------+------------------+-----------------+")
    return correction


def _get_real_space(conv, inv_diel, det_diel, latt, longest, r_c, axis, sup_latt):
    # Calculate real space component
    real_space = 0.0
    axis_ranges = [range(-a, a) for a in axis]

    # Pre-compute square of cutoff distance for cheaper comparison than
    # separation < r_c
    r_c_sq = r_c**2

    def _real_loop_function(mno):
        # Calculate the defect's fractional position in extended supercell
        d_super = np.array(mno, dtype=float) / axis
        d_super_cart = np.dot(d_super, sup_latt)

        # Test if the new atom coordinates fall within r_c, then solve
        separation_sq = np.sum(np.square(d_super_cart))
        # Take all cases within r_c except m,n,o != 0,0,0
        if separation_sq < r_c_sq and any(mno):
            mod = np.dot(d_super_cart, inv_diel)
            dot_prod = np.dot(mod, d_super_cart)
            N = np.sqrt(dot_prod)
            contribution = 1 / np.sqrt(det_diel) * erfc(conv * N) / N
            return contribution
        else:
            return 0.0

    real_space = sum(
        _real_loop_function(mno) for mno in itertools.product(*axis_ranges)
    )
    return real_space


def _get_recip(
    conv,
    inv_diel,
    det_diel,
    latt,
    recip_axis,
    recip_volume,
    recip_latt,
    dielectric_matrix,
):
    # convert factional motif to reciprocal space and
    # calculate reciprocal space supercell parallelpiped
    recip_sup_latt = np.dot(np.diag(recip_axis), recip_latt)

    # Calculate reciprocal space component
    axis_ranges = [range(-a, a) for a in recip_axis]

    def _recip_loop_function(mno):
        # Calculate the defect's fractional position in extended supercell
        d_super = np.array(mno, dtype=float) / recip_axis
        d_super_cart = np.dot(d_super, recip_sup_latt)

        if any(mno):
            mod = np.dot(d_super_cart, dielectric_matrix)
            dot_prod = np.dot(mod, d_super_cart)
            contribution = exp(-dot_prod / (4 * conv**2)) / dot_prod
            return contribution
        else:
            return 0.0

    reciprocal = sum(
        _recip_loop_function(mno) for mno in itertools.product(*axis_ranges)
    )
    scale_factor = 4 * np.pi / recip_volume
    return reciprocal * scale_factor
